# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=consider-iterating-dictionary

"""Vocabulary class used in the BERT."""
import json
import os

from ..data.transforms import SentencepieceTokenizer
from ..data.utils import count_tokens
from .vocab import Vocab

__all__ = ['BERTVocab']


UNKNOWN_TOKEN = '[UNK]'
PADDING_TOKEN = '[PAD]'
MASK_TOKEN = '[MASK]'
SEP_TOKEN = '[SEP]'
CLS_TOKEN = '[CLS]'


class BERTVocab(Vocab):
    """Specialization of gluonnlp.Vocab for BERT models.

    BERTVocab changes default token representations of unknown and other
    special tokens of gluonnlp.Vocab and adds convenience parameters to specify
    mask, sep and cls tokens typically used by Bert models.

    Parameters
    ----------
    counter : Counter or None, default None
        Counts text token frequencies in the text data. Its keys will be indexed according to
        frequency thresholds such as `max_size` and `min_freq`. Keys of `counter`,
        `unknown_token`, and values of `reserved_tokens` must be of the same hashable type.
        Examples: str, int, and tuple.
    max_size : None or int, default None
        The maximum possible number of the most frequent tokens in the keys of `counter` that can be
        indexed. Note that this argument does not count any token from `reserved_tokens`. Suppose
        that there are different keys of `counter` whose frequency are the same, if indexing all of
        them will exceed this argument value, such keys will be indexed one by one according to
        their __cmp__() order until the frequency threshold is met. If this argument is None or
        larger than its largest possible value restricted by `counter` and `reserved_tokens`, this
        argument has no effect.
    min_freq : int, default 1
        The minimum frequency required for a token in the keys of `counter` to be indexed.
    unknown_token : hashable object or None, default '[UNK]'
        The representation for any unknown token. In other words, any unknown token will be indexed
        as the same representation. If None, looking up an unknown token will result in KeyError.
    padding_token : hashable object or None, default '[PAD]'
        The representation for the special token of padding token.
    bos_token : hashable object or None, default None
        The representation for the special token of beginning-of-sequence token.
    eos_token : hashable object or None, default None
        The representation for the special token of end-of-sequence token.
    mask_token : hashable object or None, default '[MASK]'
        The representation for the special token of mask token for BERT.
    sep_token : hashable object or None, default '[SEP]'
        A token used to separate sentence pairs for BERT.
    cls_token : hashable object or None, default '[CLS]'
        Classification symbol for BERT.
    reserved_tokens : list of hashable objects or None, default None
        A list specifying additional tokens to be added to the vocabulary.
        `reserved_tokens` cannot contain `unknown_token` or duplicate reserved
        tokens.
        Keys of `counter`, `unknown_token`, and values of `reserved_tokens`
        must be of the same hashable type. Examples of hashable types are str,
        int, and tuple.
    token_to_idx : dict mapping tokens (hashable objects) to int or None, default None
        Optionally specifies the indices of tokens to be used by the
        vocabulary. Each token in `token_to_index` must be part of the Vocab
        and each index can only be associated with a single token.
        `token_to_idx` is not required to contain a mapping for all tokens. For
        example, it is valid to only set the `unknown_token` index to 10 (instead
        of the default of 0) with `token_to_idx = {'<unk>': 10}`.

    Attributes
    ----------
    embedding : instance of :class:`gluonnlp.embedding.TokenEmbedding`
        The embedding of the indexed tokens.
    idx_to_token : list of strs
        A list of indexed tokens where the list indices and the token indices are aligned.
    reserved_tokens : list of strs or None
        A list of reserved tokens that will always be indexed.
    token_to_idx : dict mapping str to int
        A dict mapping each token to its index integer.
    unknown_token : hashable object or None, default '[UNK]'
        The representation for any unknown token. In other words, any unknown token will be indexed
        as the same representation.
    padding_token : hashable object or None, default '[PAD]'
        The representation for padding token.
    bos_token : hashable object or None, default None
        The representation for beginning-of-sentence token.
    eos_token : hashable object or None, default None
        The representation for end-of-sentence token.
    mask_token : hashable object or None, default '[MASK]'
        The representation for the special token of mask token for BERT.
    sep_token : hashable object or None, default '[SEP]'
        a token used to separate sentence pairs for BERT.
    cls_token : hashable object or None, default '[CLS]'

    """

    def __init__(self, counter=None, max_size=None, min_freq=1, unknown_token=UNKNOWN_TOKEN,
                 padding_token=PADDING_TOKEN, bos_token=None, eos_token=None, mask_token=MASK_TOKEN,
                 sep_token=SEP_TOKEN, cls_token=CLS_TOKEN, reserved_tokens=None, token_to_idx=None):

        super(BERTVocab, self).__init__(counter=counter, max_size=max_size, min_freq=min_freq,
                                        unknown_token=unknown_token, padding_token=padding_token,
                                        bos_token=bos_token, eos_token=eos_token,
                                        reserved_tokens=reserved_tokens, cls_token=cls_token,
                                        sep_token=sep_token, mask_token=mask_token,
                                        token_to_idx=token_to_idx)

    @classmethod
    def from_json(cls, json_str):
        """Deserialize BERTVocab object from json string.

        Parameters
        ----------
        json_str : str
            Serialized json string of a BERTVocab object.


        Returns
        -------
        BERTVocab
        """
        vocab_dict = json.loads(json_str)
        token_to_idx = vocab_dict.get('token_to_idx')
        unknown_token = vocab_dict.get('unknown_token')
        reserved_tokens = vocab_dict.get('reserved_tokens')
        identifiers_to_tokens = vocab_dict.get('identifiers_to_tokens', dict())

        special_tokens = {unknown_token}

        # Backward compatibility for explicit serialization of padding_token,
        # bos_token, eos_token, mask_token, sep_token, cls_token handling in
        # the json string as done in older versions of GluonNLP.
        deprecated_arguments = [
            'padding_token', 'bos_token', 'eos_token', 'mask_token', 'sep_token', 'cls_token'
        ]
        for token_name in deprecated_arguments:
            token = vocab_dict.get(token_name)
            if token is not None:
                assert token_name not in identifiers_to_tokens, 'Invalid json string. ' \
                    '{} was serialized twice.'.format(token_name)
                identifiers_to_tokens[token_name] = token

        # Separate reserved from special tokens
        special_tokens.update(identifiers_to_tokens.values())
        if reserved_tokens is not None:
            reserved_tokens = [
                t for t in reserved_tokens if t not in special_tokens
            ]

        return cls(counter=count_tokens(token_to_idx.keys()),
                   unknown_token=unknown_token,
                   reserved_tokens=reserved_tokens,
                   token_to_idx=token_to_idx,
                   **identifiers_to_tokens)

    @classmethod
    def from_sentencepiece(cls,
                           path,
                           mask_token=MASK_TOKEN,
                           sep_token=SEP_TOKEN,
                           cls_token=CLS_TOKEN,
                           unknown_token=None,
                           padding_token=None,
                           bos_token=None,
                           eos_token=None,
                           reserved_tokens=None):
        """BERTVocab from pre-trained sentencepiece Tokenizer

        Parameters
        ----------
        path : str
            Path to the pre-trained subword tokenization model.
        mask_token : hashable object or None, default '[MASK]'
            The representation for the special token of mask token for BERT
        sep_token : hashable object or None, default '[SEP]'
            a token used to separate sentence pairs for BERT.
        cls_token : hashable object or None, default '[CLS]'
        unknown_token : hashable object or None, default None
            The representation for any unknown token. In other words,
            any unknown token will be indexed as the same representation.
            If set to None, it is set to the token corresponding to the unk_id()
            in the loaded sentencepiece model.
        padding_token : hashable object or None, default '[PAD]'
            The representation for padding token.
        bos_token : hashable object or None, default None
            The representation for the begin of sentence token.
            If set to None, it is set to the token corresponding to the bos_id()
            in the loaded sentencepiece model.
        eos_token : hashable object or None, default None
            The representation for the end of sentence token.
            If set to None, it is set to the token corresponding to the bos_id()
            in the loaded sentencepiece model.
        reserved_tokens : list of strs or None, optional
            A list of reserved tokens that will always be indexed.

        Returns
        -------
        BERTVocab
        """
        sp = SentencepieceTokenizer(os.path.expanduser(path))
        processor = sp._processor

        # we manually construct token_to_idx, idx_to_token and relevant fields for a BERT vocab.
        token_to_idx = {t: i for i, t in enumerate(sp.tokens)}

        def _check_consistency(processor, token_id, provided_token):
            """Check if provided_token is consistent with the special token inferred
            from the loaded sentencepiece vocab."""
            if token_id >= 0:
                # sentencepiece contains this special token.
                token = processor.IdToPiece(token_id)
                if provided_token:
                    assert provided_token == token
                provided_token = token
            return provided_token

        unknown_token = _check_consistency(processor, processor.unk_id(), unknown_token)
        bos_token = _check_consistency(processor, processor.bos_id(), bos_token)
        eos_token = _check_consistency(processor, processor.eos_id(), eos_token)
        padding_token = _check_consistency(processor, processor.pad_id(), padding_token)

        return cls(counter=count_tokens(token_to_idx.keys()),
                   unknown_token=unknown_token,
                   padding_token=padding_token,
                   bos_token=bos_token,
                   eos_token=eos_token,
                   mask_token=mask_token,
                   sep_token=sep_token,
                   cls_token=cls_token,
                   reserved_tokens=reserved_tokens,
                   token_to_idx=token_to_idx)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=
"""WikiText corpora."""

__all__ = ['WikiText2', 'WikiText103', 'WikiText2Raw', 'WikiText103Raw']

import os
import shutil
import zipfile

from mxnet.gluon.utils import _get_repo_file_url, check_sha1, download

from ... import _constants as C
from ..dataset import CorpusDataset
from ..registry import register
from ...base import get_home_dir


class _WikiText(CorpusDataset):
    def __init__(self, namespace, segment, bos, eos, flatten, skip_empty, root,
                 **kwargs):
        root = os.path.expanduser(root)
        if not os.path.isdir(root):
            os.makedirs(root)
        self._root = root
        self._namespace = 'gluon/dataset/{}'.format(namespace)
        self._segment = segment
        super(_WikiText, self).__init__(
            self._get_data(),
            bos=bos,
            eos=eos,
            flatten=flatten,
            skip_empty=skip_empty,
            **kwargs)

    def _get_data(self):
        archive_file_name, archive_hash = self._archive_file
        data_file_name, data_hash = self._data_file[self._segment]
        root = self._root
        path = os.path.join(root, data_file_name)
        if not os.path.exists(path) or not check_sha1(path, data_hash):
            downloaded_file_path = download(_get_repo_file_url(self._namespace, archive_file_name),
                                            path=root,
                                            sha1_hash=archive_hash)

            with zipfile.ZipFile(downloaded_file_path, 'r') as zf:
                for member in zf.namelist():
                    filename = os.path.basename(member)
                    if filename:
                        dest = os.path.join(root, filename)
                        with zf.open(member) as source, \
                                open(dest, 'wb') as target:
                            shutil.copyfileobj(source, target)
        return path


@register(segment=['train', 'val', 'test'])
class WikiText2(_WikiText):
    """WikiText-2 word-level dataset for language modeling, from Salesforce research.

    WikiText2 is implemented as CorpusDataset with the default flatten=True.

    From
    https://blog.einstein.ai/the-wikitext-long-term-dependency-language-modeling-dataset/

    License: Creative Commons Attribution-ShareAlike

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    flatten : bool, default True
        Whether to return all samples as flattened tokens. If True, each sample is a token.
    skip_empty : bool, default True
        Whether to skip the empty samples produced from sample_splitters. If False, `bos` and `eos`
        will be added in empty samples.
    tokenizer : function, default str.split
        A function that splits each sample string into list of tokens.
    bos : str or None, default None
        The token to add at the beginning of each sentence. If None, nothing is added.
    eos : str or None, default '<eos>'
        The token to add at the end of each sentence. If None, nothing is added.
    root : str, default '$MXNET_HOME/datasets/wikitext-2'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> wikitext2 = gluonnlp.data.WikiText2('val', root='./datasets/wikitext2')
    -etc-
    >>> len(wikitext2)
    216347
    >>> wikitext2[0]
    '='
    >>> wikitext2 = gluonnlp.data.WikiText2('val', flatten=False,
    ...                                     root='./datasets/wikitext2')
    >>> len(wikitext2)
    2461
    >>> wikitext2[0]
    ['=', 'Homarus', 'gammarus', '=', '<eos>']
    >>> wikitext2 = gluonnlp.data.WikiText2('val', flatten=False, bos='<bos>', eos=None,
    ...                                     root='./datasets/wikitext2')
    >>> wikitext2[0]
    ['<bos>', '=', 'Homarus', 'gammarus', '=']
    >>> wikitext2 = gluonnlp.data.WikiText2('val', flatten=False, bos='<bos>', eos=None,
    ...                                     skip_empty=False, root='./datasets/wikitext2')
    >>> len(wikitext2)
    3760
    >>> wikitext2[0]
    ['<bos>']
    """

    def __init__(self,
                 segment='train',
                 flatten=True,
                 skip_empty=True,
                 tokenizer=lambda s: s.split(),
                 bos=None,
                 eos=C.EOS_TOKEN,
                 root=os.path.join(get_home_dir(), 'datasets', 'wikitext-2'),
                 **kwargs):
        self._archive_file = ('wikitext-2-v1.zip',
                              '3c914d17d80b1459be871a5039ac23e752a53cbe')
        self._data_file = {
            'train': ('wiki.train.tokens',
                      '863f29c46ef9d167fff4940ec821195882fe29d1'),
            'val': ('wiki.valid.tokens',
                    '0418625c8b4da6e4b5c7a0b9e78d4ae8f7ee5422'),
            'test': ('wiki.test.tokens',
                     'c7b8ce0aa086fb34dab808c5c49224211eb2b172')
        }
        super(WikiText2, self).__init__(
            'wikitext-2',
            segment=segment,
            bos=bos,
            eos=eos,
            flatten=flatten,
            skip_empty=skip_empty,
            root=root,
            tokenizer=tokenizer,
            **kwargs)


@register(segment=['train', 'val', 'test'])
class WikiText103(_WikiText):
    """WikiText-103 word-level dataset for language modeling, from Salesforce research.

    WikiText103 is implemented as CorpusDataset with the default flatten=True.

    From
    https://blog.einstein.ai/the-wikitext-long-term-dependency-language-modeling-dataset/

    License: Creative Commons Attribution-ShareAlike

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    flatten : bool, default True
        Whether to return all samples as flattened tokens. If True, each sample is a token.
    skip_empty : bool, default True
        Whether to skip the empty samples produced from sample_splitters. If False, `bos` and `eos`
        will be added in empty samples.
    tokenizer : function, default str.split
        A function that splits each sample string into list of tokens.
    bos : str or None, default None
        The token to add at the beginning of each sentence. If None, nothing is added.
    eos : str or None, default '<eos>'
        The token to add at the end of each sentence. If None, nothing is added.
    root : str, default '$MXNET_HOME/datasets/wikitext-103'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> wikitext103 = gluonnlp.data.WikiText103('val', root='./datasets/wikitext103')
    -etc-
    >>> len(wikitext103)
    216347
    >>> wikitext103[0]
    '='
    >>> wikitext103 = gluonnlp.data.WikiText103('val', flatten=False,
    ...                                         root='./datasets/wikitext103')
    >>> len(wikitext103)
    2461
    >>> wikitext103[0]
    ['=', 'Homarus', 'gammarus', '=', '<eos>']
    >>> wikitext103 = gluonnlp.data.WikiText103('val', flatten=False, bos='<bos>', eos=None,
    ...                                         root='./datasets/wikitext103')
    >>> wikitext103[0]
    ['<bos>', '=', 'Homarus', 'gammarus', '=']
    >>> wikitext103 = gluonnlp.data.WikiText103('val', flatten=False, bos='<bos>', eos=None,
    ...                                         skip_empty=False, root='./datasets/wikitext103')
    >>> len(wikitext103)
    3760
    >>> wikitext103[0]
    ['<bos>']
    """

    def __init__(self,
                 segment='train',
                 flatten=True,
                 skip_empty=True,
                 tokenizer=lambda s: s.split(),
                 bos=None,
                 eos=C.EOS_TOKEN,
                 root=os.path.join(get_home_dir(), 'datasets',
                                   'wikitext-103'),
                 **kwargs):
        self._archive_file = ('wikitext-103-v1.zip',
                              '0aec09a7537b58d4bb65362fee27650eeaba625a')
        self._data_file = {
            'train': ('wiki.train.tokens',
                      'b7497e2dfe77e72cfef5e3dbc61b7b53712ac211'),
            'val': ('wiki.valid.tokens',
                    'c326ac59dc587676d58c422eb8a03e119582f92b'),
            'test': ('wiki.test.tokens',
                     '8a5befc548865cec54ed4273cf87dbbad60d1e47')
        }
        super(WikiText103, self).__init__(
            'wikitext-103',
            segment=segment,
            bos=bos,
            eos=eos,
            flatten=flatten,
            skip_empty=skip_empty,
            root=root,
            tokenizer=tokenizer,
            **kwargs)


@register(segment=['train', 'val', 'test'])
class WikiText2Raw(_WikiText):
    """WikiText-2 character-level dataset for language modeling

    WikiText2Raw is implemented as CorpusDataset with the default flatten=True.

    From Salesforce research:
    https://blog.einstein.ai/the-wikitext-long-term-dependency-language-modeling-dataset/

    License: Creative Commons Attribution-ShareAlike

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    flatten : bool, default True
        Whether to return all samples as flattened tokens. If True, each sample is a token.
    skip_empty : bool, default True
        Whether to skip the empty samples produced from sample_splitters. If False, `bos` and `eos`
        will be added in empty samples.
    tokenizer : function, default s.encode('utf-8')
        A function that splits each sample string into list of tokens.
        The tokenizer can also be used to convert everything to lowercase.
        E.g. with tokenizer=lambda s: s.lower().encode('utf-8')
    bos : str or None, default None
        The token to add at the beginning of each sentence. If None, nothing is added.
    eos : str or None, default '<eos>'
        The token to add at the end of each sentence. If None, nothing is added.
    root : str, default '$MXNET_HOME/datasets/wikitext-2'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> wikitext2 = gluonnlp.data.WikiText2Raw('val', root='./datasets/wikitext2')
    -etc-
    >>> len(wikitext2)
    1136862
    >>> wikitext2[0]
    61
    >>> type(wikitext2[0])
    <class 'int'>
    >>> wikitext2 = gluonnlp.data.WikiText2Raw('val', flatten=False,
    ...                                        tokenizer=None, root='./datasets/wikitext2')
    >>> len(wikitext2)
    2461
    >>> wikitext2[0]
    '= Homarus gammarus ='
    >>> wikitext2 = gluonnlp.data.WikiText2Raw('val', flatten=False, bos='<bos>', eos=None,
    ...                                        tokenizer=lambda s: s.split(),
    ...                                        root='./datasets/wikitext2')
    >>> wikitext2[0]
    ['<bos>', '=', 'Homarus', 'gammarus', '=']
    """

    def __init__(self,
                 segment='train',
                 flatten=True,
                 skip_empty=True,
                 bos=None,
                 eos=None,
                 tokenizer=lambda s: s.encode('utf-8'),
                 root=os.path.join(get_home_dir(), 'datasets', 'wikitext-2'),
                 **kwargs):
        self._archive_file = ('wikitext-2-raw-v1.zip',
                              '3b6993c138fc61c95f7fffd900fef68f8411371d')
        self._data_file = {
            'train': ('wiki.train.raw',
                      'd33faf256327882db0edc7c67cd098d1051a2112'),
            'val': ('wiki.valid.raw',
                    'db78d4db83700cba1b1bf4a9381087043db2876d'),
            'test': ('wiki.test.raw',
                     '6f1fe2054a940eebfc76b284b09680763b37f5ea')
        }

        super(WikiText2Raw, self).__init__(
            'wikitext-2',
            segment=segment,
            bos=bos,
            eos=eos,
            flatten=flatten,
            skip_empty=skip_empty,
            root=root,
            tokenizer=tokenizer,
            **kwargs)


@register(segment=['train', 'val', 'test'])
class WikiText103Raw(_WikiText):
    """WikiText-103 character-level dataset for language modeling

    WikiText103Raw is implemented as CorpusDataset with the default flatten=True.

    From Salesforce research:
    https://blog.einstein.ai/the-wikitext-long-term-dependency-language-modeling-dataset/

    License: Creative Commons Attribution-ShareAlike

    Parameters
    ----------
    segment : {'train', 'val', 'test'}, default 'train'
        Dataset segment.
    flatten : bool, default True
        Whether to return all samples as flattened tokens. If True, each sample is a token.
    skip_empty : bool, default True
        Whether to skip the empty samples produced from sample_splitters. If False, `bos` and `eos`
        will be added in empty samples.
    tokenizer : function, default s.encode('utf-8')
        A function that splits each sample string into list of tokens.
        The tokenizer can also be used to convert everything to lowercase.
        E.g. with tokenizer=lambda s: s.lower().encode('utf-8')
    bos : str or None, default None
        The token to add at the beginning of each sentence. If None, nothing is added.
    eos : str or None, default '<eos>'
        The token to add at the end of each sentence. If None, nothing is added.
    root : str, default '$MXNET_HOME/datasets/wikitext-103'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> wikitext103 = gluonnlp.data.WikiText103Raw('val', root='./datasets/wikitext103')
    -etc-
    >>> len(wikitext103)
    1136862
    >>> wikitext103[0]
    61
    >>> wikitext103 = gluonnlp.data.WikiText103Raw('val', flatten=False,
    ...                                            root='./datasets/wikitext103')
    >>> len(wikitext103)
    2461
    >>> wikitext103[0]
    [61, 32, 72, 111, 109, 97, 114, 117, 115, 32, 103, 97, 109, 109, 97, 114, 117, 115, 32, 61]
    >>> wikitext103 = gluonnlp.data.WikiText103Raw('val', flatten=False, tokenizer=None,
    ...                                            root='./datasets/wikitext103')
    >>> wikitext103[0]
    '= Homarus gammarus ='
    """

    def __init__(self,
                 segment='train',
                 flatten=True,
                 skip_empty=True,
                 tokenizer=lambda s: s.encode('utf-8'),
                 bos=None,
                 eos=None,
                 root=os.path.join(get_home_dir(), 'datasets',
                                   'wikitext-103'),
                 **kwargs):
        self._archive_file = ('wikitext-103-raw-v1.zip',
                              '86f2375181b9247049d9c9205fad2b71b274b568')
        self._data_file = {
            'train': ('wiki.train.raw',
                      '3d06627c15e834408cfee91293f862c11c1cc9ef'),
            'val': ('wiki.valid.raw',
                    'db78d4db83700cba1b1bf4a9381087043db2876d'),
            'test': ('wiki.test.raw',
                     '6f1fe2054a940eebfc76b284b09680763b37f5ea')
        }
        super(WikiText103Raw, self).__init__(
            'wikitext-103',
            segment=segment,
            bos=bos,
            eos=eos,
            flatten=flatten,
            skip_empty=skip_empty,
            root=root,
            tokenizer=tokenizer,
            **kwargs)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Bidirectional LM encoder."""
__all__ = ['BiLMEncoder']

from mxnet import gluon
from mxnet.gluon import rnn
from .utils import _get_rnn_cell


class BiLMEncoder(gluon.HybridBlock):
    r"""Bidirectional LM encoder.

    We implement the encoder of the biLM proposed in the following work::

        @inproceedings{Peters:2018,
        author={Peters, Matthew E. and  Neumann, Mark and Iyyer, Mohit and Gardner, Matt and Clark,
        Christopher and Lee, Kenton and Zettlemoyer, Luke},
        title={Deep contextualized word representations},
        booktitle={Proc. of NAACL},
        year={2018}
        }

    Parameters
    ----------
    mode : str
        The type of RNN cell to use. Options are 'lstmpc', 'rnn_tanh', 'rnn_relu', 'lstm', 'gru'.
    num_layers : int
        The number of RNN cells in the encoder.
    input_size : int
        The initial input size of in the RNN cell.
    hidden_size : int
        The hidden size of the RNN cell.
    dropout : float
        The dropout rate to use for encoder output.
    skip_connection : bool
        Whether to add skip connections (add RNN cell input to output)
    proj_size : int
        The projection size of each LSTMPCellWithClip cell
    cell_clip : float
        Clip cell state between [-cellclip, cell_clip] in LSTMPCellWithClip cell
    proj_clip : float
        Clip projection between [-projclip, projclip] in LSTMPCellWithClip cell
    """

    def __init__(self, mode, num_layers, input_size, hidden_size, dropout=0.0,
                 skip_connection=True, proj_size=None, cell_clip=None, proj_clip=None, **kwargs):
        super(BiLMEncoder, self).__init__(**kwargs)

        self._mode = mode
        self._num_layers = num_layers
        self._input_size = input_size
        self._hidden_size = hidden_size
        self._dropout = dropout
        self._skip_connection = skip_connection
        self._proj_size = proj_size
        self._cell_clip = cell_clip
        self._proj_clip = proj_clip

        with self.name_scope():
            lstm_input_size = self._input_size
            self.forward_layers = rnn.HybridSequentialRNNCell()
            with self.forward_layers.name_scope():
                for layer_index in range(self._num_layers):
                    forward_layer = _get_rnn_cell(mode=self._mode,
                                                  num_layers=1,
                                                  input_size=lstm_input_size,
                                                  hidden_size=self._hidden_size,
                                                  dropout=0
                                                  if layer_index == num_layers - 1
                                                  else self._dropout,
                                                  weight_dropout=0,
                                                  var_drop_in=0,
                                                  var_drop_state=0,
                                                  var_drop_out=0,
                                                  skip_connection=False
                                                  if layer_index == 0
                                                  else self._skip_connection,
                                                  proj_size=self._proj_size,
                                                  cell_clip=self._cell_clip,
                                                  proj_clip=self._proj_clip)

                    self.forward_layers.add(forward_layer)
                    lstm_input_size = self._hidden_size \
                        if self._proj_size is None else self._proj_size

            lstm_input_size = self._input_size
            self.backward_layers = rnn.HybridSequentialRNNCell()
            with self.backward_layers.name_scope():
                for layer_index in range(self._num_layers):
                    backward_layer = _get_rnn_cell(mode=self._mode,
                                                   num_layers=1,
                                                   input_size=lstm_input_size,
                                                   hidden_size=self._hidden_size,
                                                   dropout=0
                                                   if layer_index == num_layers - 1
                                                   else self._dropout,
                                                   weight_dropout=0,
                                                   var_drop_in=0,
                                                   var_drop_state=0,
                                                   var_drop_out=0,
                                                   skip_connection=False
                                                   if layer_index == 0
                                                   else self._skip_connection,
                                                   proj_size=self._proj_size,
                                                   cell_clip=self._cell_clip,
                                                   proj_clip=self._proj_clip)
                    self.backward_layers.add(backward_layer)
                    lstm_input_size = self._hidden_size \
                        if self._proj_size is None else self._proj_size

    def begin_state(self, func, **kwargs):
        return [self.forward_layers[0][0].begin_state(func=func, **kwargs)
                for _ in range(self._num_layers)], \
               [self.backward_layers[0][0].begin_state(func=func, **kwargs)
                for _ in range(self._num_layers)]

    def hybrid_forward(self, F, inputs, states=None, mask=None):
        # pylint: disable=arguments-differ
        # pylint: disable=unused-argument
        """Defines the forward computation for cache cell. Arguments can be either
        :py:class:`NDArray` or :py:class:`Symbol`.

        Parameters
        ----------
        inputs : NDArray
            The input data layout='TNC'.
        states : Tuple[List[List[NDArray]]]
            The states. including:
            states[0] indicates the states used in forward layer,
            Each layer has a list of two initial tensors with
            shape (batch_size, proj_size) and (batch_size, hidden_size).
            states[1] indicates the states used in backward layer,
            Each layer has a list of two initial tensors with
            shape (batch_size, proj_size) and (batch_size, hidden_size).

        Returns
        --------
        out: NDArray
            The output data with shape (num_layers, seq_len, batch_size, 2*input_size).
        [states_forward, states_backward] : List
            Including:
            states_forward: The out states from forward layer,
            which has the same structure with *states[0]*.
            states_backward: The out states from backward layer,
            which has the same structure with *states[1]*.
        """
        states_forward, states_backward = states
        if mask is not None:
            sequence_length = mask.sum(axis=1)

        outputs_forward = []
        outputs_backward = []

        for layer_index in range(self._num_layers):
            if layer_index == 0:
                layer_inputs = inputs
            else:
                layer_inputs = outputs_forward[layer_index - 1]
            output, states_forward[layer_index] = F.contrib.foreach(
                self.forward_layers[layer_index],
                layer_inputs,
                states_forward[layer_index])
            outputs_forward.append(output)

            if layer_index == 0:
                layer_inputs = inputs
            else:
                layer_inputs = outputs_backward[layer_index - 1]

            if mask is not None:
                layer_inputs = F.SequenceReverse(layer_inputs,
                                                 sequence_length=sequence_length,
                                                 use_sequence_length=True, axis=0)
            else:
                layer_inputs = F.SequenceReverse(layer_inputs, axis=0)
            output, states_backward[layer_index] = F.contrib.foreach(
                self.backward_layers[layer_index],
                layer_inputs,
                states_backward[layer_index])
            if mask is not None:
                backward_out = F.SequenceReverse(output,
                                                 sequence_length=sequence_length,
                                                 use_sequence_length=True, axis=0)
            else:
                backward_out = F.SequenceReverse(output, axis=0)
            outputs_backward.append(backward_out)
        out = F.concat(*[F.stack(*outputs_forward, axis=0),
                         F.stack(*outputs_backward, axis=0)], dim=-1)

        return out, [states_forward, states_backward]



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
# pylint: disable=eval-used, redefined-outer-name
"""Models for intrinsic and extrinsic word embedding evaluation"""

import mxnet as mx
from mxnet import registry
from mxnet.gluon import HybridBlock

__all__ = [
    'register', 'create', 'list_evaluation_functions',
    'WordEmbeddingSimilarityFunction', 'WordEmbeddingAnalogyFunction',
    'CosineSimilarity', 'ThreeCosMul', 'ThreeCosAdd',
    'WordEmbeddingSimilarity', 'WordEmbeddingAnalogy']


class _WordEmbeddingEvaluationFunction(HybridBlock):  # pylint: disable=abstract-method
    """Base class for word embedding evaluation functions."""


class WordEmbeddingSimilarityFunction(_WordEmbeddingEvaluationFunction):  # pylint: disable=abstract-method
    """Base class for word embedding similarity functions."""


class WordEmbeddingAnalogyFunction(_WordEmbeddingEvaluationFunction):  # pylint: disable=abstract-method
    """Base class for word embedding analogy functions.

    Parameters
    ----------
    idx_to_vec : mxnet.ndarray.NDArray
        Embedding matrix.
    k : int, default 1
        Number of analogies to predict per input triple.
    eps : float, optional, default=1e-10
        A small constant for numerical stability.
    """


###############################################################################
# Similarity and analogy functions registry helpers
###############################################################################
_REGSITRY_KIND_CLASS_MAP = {
    'similarity': WordEmbeddingSimilarityFunction,
    'analogy': WordEmbeddingAnalogyFunction
}


def register(class_):
    """Registers a new word embedding evaluation function.

    Once registered, we can create an instance with
    :func:`~gluonnlp.embedding.evaluation.create`.

    Examples
    --------
    >>> @gluonnlp.embedding.evaluation.register
    ... class MySimilarityFunction(gluonnlp.embedding.evaluation.WordEmbeddingSimilarityFunction):
    ...     def __init__(self, eps=1e-10):
    ...         pass
    >>> similarity_function = gluonnlp.embedding.evaluation.create('similarity',
    ...                                                            'MySimilarityFunction')
    >>> print(type(similarity_function))
    <class 'gluonnlp.embedding.evaluation.MySimilarityFunction'>

    >>> @gluonnlp.embedding.evaluation.register
    ... class MyAnalogyFunction(gluonnlp.embedding.evaluation.WordEmbeddingAnalogyFunction):
    ...     def __init__(self, k=1, eps=1E-10):
    ...         pass
    >>> analogy_function = gluonnlp.embedding.evaluation.create('analogy', 'MyAnalogyFunction')
    >>> print(type(analogy_function))
    <class 'gluonnlp.embedding.evaluation.MyAnalogyFunction'>

    """

    if issubclass(class_, WordEmbeddingSimilarityFunction):
        register_ = registry.get_register_func(
            WordEmbeddingSimilarityFunction,
            'word embedding similarity evaluation function')
    elif issubclass(class_, WordEmbeddingAnalogyFunction):
        register_ = registry.get_register_func(
            WordEmbeddingAnalogyFunction,
            'word embedding analogy evaluation function')
    else:
        raise RuntimeError(
            'The custom function must either subclass '
            'WordEmbeddingSimilarityFunction or WordEmbeddingAnalogyFunction')

    return register_(class_)


def create(kind, name, **kwargs):
    """Creates an instance of a registered word embedding evaluation function.

    Parameters
    ----------
    kind : ['similarity', 'analogy']
        Return only valid names for similarity, analogy or both kinds of
        functions.
    name : str
        The evaluation function name (case-insensitive).


    Returns
    -------
    An instance of
    :class:`gluonnlp.embedding.evaluation.WordEmbeddingAnalogyFunction`:
    or
    :class:`gluonnlp.embedding.evaluation.WordEmbeddingSimilarityFunction`:
        An instance of the specified evaluation function.

    """
    if kind not in _REGSITRY_KIND_CLASS_MAP.keys():
        raise KeyError(
            'Cannot find `kind` {}. Use '
            '`list_evaluation_functions(kind=None).keys()` to get'
            'all the valid kinds of evaluation functions.'.format(kind))

    create_ = registry.get_create_func(
        _REGSITRY_KIND_CLASS_MAP[kind],
        'word embedding {} evaluation function'.format(kind))

    return create_(name, **kwargs)


def list_evaluation_functions(kind=None):
    """Get valid word embedding functions names.

    Parameters
    ----------
    kind : ['similarity', 'analogy', None]
        Return only valid names for similarity, analogy or both kinds of functions.

    Returns
    -------
    dict or list:
        A list of all the valid evaluation function names for the specified
        kind. If kind is set to None, returns a dict mapping each valid name to
        its respective output list. The valid names can be plugged in
        `gluonnlp.model.word_evaluation_model.create(name)`.

    """

    if kind is None:
        kind = tuple(_REGSITRY_KIND_CLASS_MAP.keys())

    if not isinstance(kind, tuple):
        if kind not in _REGSITRY_KIND_CLASS_MAP.keys():
            raise KeyError(
                'Cannot find `kind` {}. Use '
                '`list_evaluation_functions(kind=None).keys()` to get all the'
                'valid kinds of evaluation functions.'.format(kind))

        reg = registry.get_registry(_REGSITRY_KIND_CLASS_MAP[kind])
        return list(reg.keys())
    else:
        return {name: list_evaluation_functions(kind=name) for name in kind}


###############################################################################
# Word embedding similarity functions
###############################################################################
@register
class CosineSimilarity(WordEmbeddingSimilarityFunction):
    """Computes the cosine similarity.

    Parameters
    ----------
    eps : float, optional, default=1e-10
        A small constant for numerical stability.

    """

    def __init__(self, eps=1e-10, **kwargs):
        super(CosineSimilarity, self).__init__(**kwargs)
        self.eps = eps

    def hybrid_forward(self, F, x, y):  # pylint: disable=arguments-differ
        """Compute the cosine similarity between two batches of vectors.

        The cosine similarity is the dot product between the L2 normalized
        vectors.

        Parameters
        ----------
        x : Symbol or NDArray
        y : Symbol or NDArray

        Returns
        -------
        similarity : Symbol or NDArray
            The similarity computed by WordEmbeddingSimilarity.similarity_function.

        """

        x = F.L2Normalization(x, eps=self.eps)
        y = F.L2Normalization(y, eps=self.eps)
        x = F.expand_dims(x, axis=1)
        y = F.expand_dims(y, axis=2)
        return F.batch_dot(x, y).reshape((-1, ))


###############################################################################
# Word embedding analogy functions
###############################################################################
@register
class ThreeCosMul(WordEmbeddingAnalogyFunction):
    """The 3CosMul analogy function.

    The 3CosMul analogy function is defined as

    .. math::
        \\arg\\max_{b^* ∈ V}\\frac{\\cos(b^∗, b) \\cos(b^*, a)}{cos(b^*, a^*) + ε}

    See the following paper for more details:

    - Levy, O., & Goldberg, Y. (2014). Linguistic regularities in sparse and
      explicit word representations. In R. Morante, & W. Yih, Proceedings of the
      Eighteenth Conference on Computational Natural Language Learning, CoNLL 2014,
      Baltimore, Maryland, USA, June 26-27, 2014 (pp. 171–180). : ACL.

    Parameters
    ----------
    idx_to_vec : mxnet.ndarray.NDArray
        Embedding matrix.
    k : int, default 1
        Number of analogies to predict per input triple.
    exclude_question_words : bool, default True
        Exclude the 3 question words from being a valid answer.
    eps : float, optional, default=1e-10
        A small constant for numerical stability.

    """

    def __init__(self, idx_to_vec, k=1, eps=1E-10, exclude_question_words=True, **kwargs):
        super(ThreeCosMul, self).__init__(**kwargs)

        self.k = k
        self.eps = eps
        self._exclude_question_words = exclude_question_words

        self._vocab_size, self._embed_size = idx_to_vec.shape

        idx_to_vec = mx.nd.L2Normalization(idx_to_vec, eps=self.eps)
        with self.name_scope():
            self.weight = self.params.get_constant('weight', idx_to_vec)

    def hybrid_forward(self, F, words1, words2, words3, weight):  # pylint: disable=arguments-differ
        """Compute ThreeCosMul for given question words.

        Parameters
        ----------
        words1 : Symbol or NDArray
            Question words at first position. Shape (batch_size, )
        words2 : Symbol or NDArray
            Question words at second position. Shape (batch_size, )
        words3 : Symbol or NDArray
            Question words at third position. Shape (batch_size, )

        Returns
        -------
        Symbol or NDArray
            Predicted answer words. Shape (batch_size, k).

        """
        words123 = F.concat(words1, words2, words3, dim=0)
        embeddings_words123 = F.Embedding(words123, weight,
                                          input_dim=self._vocab_size,
                                          output_dim=self._embed_size)
        similarities = F.FullyConnected(
            embeddings_words123, weight, no_bias=True,
            num_hidden=self._vocab_size, flatten=False)
        # Map cosine similarities to [0, 1]
        similarities = (similarities + 1) / 2

        sim_w1w4, sim_w2w4, sim_w3w4 = F.split(similarities, num_outputs=3,
                                               axis=0)

        sim = (sim_w2w4 * sim_w3w4) / (sim_w1w4 + self.eps)

        if self._exclude_question_words:
            for words in [words1, words2, words3]:
                sim = sim * F.one_hot(words, self.weight.shape[0], 0, 1)

        pred_idxs = F.topk(sim, k=self.k)
        return pred_idxs


@register
class ThreeCosAdd(WordEmbeddingAnalogyFunction):
    """The 3CosAdd analogy function.

    The 3CosAdd analogy function is defined as

    .. math::
        \\arg\\max_{b^* ∈ V}[\\cos(b^∗, b - a + a^*)]

    See the following paper for more details:

    - Levy, O., & Goldberg, Y. (2014). Linguistic regularities in sparse and
      explicit word representations. In R. Morante, & W. Yih, Proceedings of the
      Eighteenth Conference on Computational Natural Language Learning, CoNLL 2014,
      Baltimore, Maryland, USA, June 26-27, 2014 (pp. 171–180). : ACL.

    Parameters
    ----------
    idx_to_vec : mxnet.ndarray.NDArray
        Embedding matrix.
    normalize : bool, default True
        Normalize all word embeddings before computing the analogy.
    k : int, default 1
        Number of analogies to predict per input triple.
    exclude_question_words : bool, default True
        Exclude the 3 question words from being a valid answer.
    eps : float, optional, default=1e-10
        A small constant for numerical stability.


    """

    def __init__(self,
                 idx_to_vec,
                 normalize=True,
                 k=1,
                 eps=1E-10,
                 exclude_question_words=True,
                 **kwargs):
        super(ThreeCosAdd, self).__init__(**kwargs)

        self.k = k
        self.eps = eps
        self.normalize = normalize
        self._exclude_question_words = exclude_question_words
        self._vocab_size, self._embed_size = idx_to_vec.shape

        if self.normalize:
            idx_to_vec = mx.nd.L2Normalization(idx_to_vec, eps=self.eps)

        with self.name_scope():
            self.weight = self.params.get_constant('weight', idx_to_vec)

    def hybrid_forward(self, F, words1, words2, words3, weight):  # pylint: disable=arguments-differ
        """Compute ThreeCosAdd for given question words.

        Parameters
        ----------
        words1 : Symbol or NDArray
            Question words at first position. Shape (batch_size, )
        words2 : Symbol or NDArray
            Question words at second position. Shape (batch_size, )
        words3 : Symbol or NDArray
            Question words at third position. Shape (batch_size, )

        Returns
        -------
        Symbol or NDArray
            Predicted answer words. Shape (batch_size, k).

        """
        words123 = F.concat(words1, words2, words3, dim=0)
        embeddings_words123 = F.Embedding(words123, weight,
                                          input_dim=self._vocab_size,
                                          output_dim=self._embed_size)
        if self.normalize:
            similarities = F.FullyConnected(
                embeddings_words123, weight, no_bias=True,
                num_hidden=self._vocab_size, flatten=False)
            sim_w1w4, sim_w2w4, sim_w3w4 = F.split(similarities, num_outputs=3,
                                                   axis=0)
            pred = sim_w3w4 - sim_w1w4 + sim_w2w4
        else:
            embeddings_word1, embeddings_word2, embeddings_word3 = F.split(
                embeddings_words123, num_outputs=3, axis=0)
            vector = (embeddings_word3 - embeddings_word1 + embeddings_word2)
            pred = F.FullyConnected(
                vector, weight, no_bias=True, num_hidden=self._vocab_size,
                flatten=False)

        if self._exclude_question_words:
            for words in [words1, words2, words3]:
                pred = pred * F.one_hot(words, self.weight.shape[0], 0, 1)

        pred_idxs = F.topk(pred, k=self.k)
        return pred_idxs


###############################################################################
# Evaluation blocks
###############################################################################
class WordEmbeddingSimilarity(HybridBlock):
    """Word embeddings similarity task evaluator.

    Parameters
    ----------
    idx_to_vec : mxnet.ndarray.NDArray
        Embedding matrix.
    similarity_function : str, default 'CosineSimilarity'
        Name of a registered WordEmbeddingSimilarityFunction.
    eps : float, optional, default=1e-10
        A small constant for numerical stability.

    """

    def __init__(self, idx_to_vec, similarity_function='CosineSimilarity',
                 eps=1e-10, **kwargs):
        super(WordEmbeddingSimilarity, self).__init__(**kwargs)

        self.eps = eps
        self._vocab_size, self._embed_size = idx_to_vec.shape

        with self.name_scope():
            self.weight = self.params.get_constant('weight', idx_to_vec)
            self.similarity = create(kind='similarity',
                                     name=similarity_function, eps=self.eps)

        if not isinstance(self.similarity, WordEmbeddingSimilarityFunction):
            raise RuntimeError(
                '{} is not a WordEmbeddingAnalogyFunction'.format(
                    self.similarity.__class__.__name__))

    def hybrid_forward(self, F, words1, words2, weight):  # pylint: disable=arguments-differ
        """Predict the similarity of words1 and words2.

        Parameters
        ----------
        words1 : Symbol or NDArray
            The indices of the words the we wish to compare to the words in words2.
        words2 : Symbol or NDArray
            The indices of the words the we wish to compare to the words in words1.

        Returns
        -------
        similarity : Symbol or NDArray
            The similarity computed by WordEmbeddingSimilarity.similarity_function.
        """
        embeddings_words1 = F.Embedding(words1, weight,
                                        input_dim=self._vocab_size,
                                        output_dim=self._embed_size)
        embeddings_words2 = F.Embedding(words2, weight,
                                        input_dim=self._vocab_size,
                                        output_dim=self._embed_size)
        similarity = self.similarity(embeddings_words1, embeddings_words2)
        return similarity


class WordEmbeddingAnalogy(HybridBlock):
    """Word embeddings analogy task evaluator.

    Parameters
    ----------
    idx_to_vec : mxnet.ndarray.NDArray
        Embedding matrix.
    analogy_function : str, default 'ThreeCosMul'
        Name of a registered WordEmbeddingAnalogyFunction.
    k : int, default 1
        Number of analogies to predict per input triple.
    exclude_question_words : bool, default True
        Exclude the 3 question words from being a valid answer.

    """

    def __init__(self, idx_to_vec, analogy_function='ThreeCosMul', k=1,
                 exclude_question_words=True, **kwargs):
        super(WordEmbeddingAnalogy, self).__init__(**kwargs)

        assert k >= 1
        self.k = k
        self.exclude_question_words = exclude_question_words

        with self.name_scope():
            self.analogy = create(
                kind='analogy',
                name=analogy_function,
                idx_to_vec=idx_to_vec,
                k=self.k,
                exclude_question_words=exclude_question_words)

        if not isinstance(self.analogy, WordEmbeddingAnalogyFunction):
            raise RuntimeError(
                '{} is not a WordEmbeddingAnalogyFunction'.format(
                    self.analogy.__class__.__name__))

    def hybrid_forward(self, F, words1, words2, words3):  # pylint: disable=arguments-differ, unused-argument
        """Compute analogies for given question words.

        Parameters
        ----------
        words1 : Symbol or NDArray
            Word indices of first question words. Shape (batch_size, ).
        words2 : Symbol or NDArray
            Word indices of second question words. Shape (batch_size, ).
        words3 : Symbol or NDArray
            Word indices of third question words. Shape (batch_size, ).

        Returns
        -------
        predicted_indices : Symbol or NDArray
            Indices of predicted analogies of shape (batch_size, k)
        """
        return self.analogy(words1, words2, words3)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=undefined-all-variable
"""Data transforms useful for language models."""

__all__ = ['CorpusBatchify', 'CorpusBPTTBatchify', 'StreamBPTTBatchify']

import itertools
import math

import numpy as np
import mxnet as mx
from mxnet.gluon.data import RandomSampler, SequentialSampler, SimpleDataset

from ..utils import slice_sequence, _slice_pad_length
from ..stream import DataStream

class CorpusBatchify:
    """Transform the dataset into N independent sequences, where N is the batch size.

    Parameters
    ----------
    vocab : gluonnlp.Vocab
        The vocabulary to use for numericalizing the dataset. Each token will be mapped to the
        index according to the vocabulary.
    batch_size : int
        The number of samples in each batch.
    """

    def __init__(self, vocab, batch_size):
        self._vocab = vocab
        self._batch_size = batch_size

    def __call__(self, data):
        """Batchify a dataset.

        Parameters
        ----------
        data : mxnet.gluon.data.Dataset
            A flat dataset to be batchified.

        Returns
        -------
        mxnet.gluon.data.Dataset
            NDArray of shape (len(data) // N, N) where N is the batch_size
            wrapped by a mxnet.gluon.data.SimpleDataset. Excessive tokens that
            don't align along the batches are discarded.
        """
        sample_len = len(data) // self._batch_size
        return SimpleDataset(
            mx.nd.array(
                self._vocab[data[:sample_len * self._batch_size]]).reshape(
                    self._batch_size, -1).T)


class CorpusBPTTBatchify:
    """Transform the dataset into batches of numericalized samples, in the way
    that the recurrent states from last batch connects with the current batch
    for each sample.

    Each sample is of shape `(seq_len, batch_size)`. When `last_batch='keep'`, the first
    dimension of last sample may be shorter than `seq_len`.

    Parameters
    ----------
    vocab : gluonnlp.Vocab
        The vocabulary to use for numericalizing the dataset. Each token will be mapped to the
        index according to the vocabulary.
    seq_len : int
        The length of each of the samples for truncated back-propagation-through-time (TBPTT).
    batch_size : int
        The number of samples in each batch.
    last_batch : {'keep', 'discard'}
        How to handle the last batch if the remaining length is less than `seq_len`.

        - keep: A batch with less samples than previous batches is returned. vocab.padding_token
          is used to pad the last batch based on batch size.

        - discard: The last batch is discarded if it's smaller than `(seq_len, batch_size)`.
    """

    def __init__(self,
                 vocab,
                 seq_len,
                 batch_size,
                 last_batch='keep'):
        self._vocab = vocab
        self._seq_len = seq_len
        self._batch_size = batch_size
        self._last_batch = last_batch

        if last_batch not in ['keep', 'discard']:
            raise ValueError(
                'Got invalid last_batch: "{}". Must be "keep" or "discard".'.
                format(last_batch))

        if self._last_batch == 'keep':
            if not self._vocab.padding_token:
                raise ValueError('vocab.padding_token must be specified '
                                 'in vocab when last_batch="keep".')

    def __call__(self, corpus):
        """Batchify a dataset.

        Parameters
        ----------
        corpus : mxnet.gluon.data.Dataset
            A flat dataset to be batchified.

        Returns
        -------
        mxnet.gluon.data.Dataset
            Batches of numericalized samples such that the recurrent states
            from last batch connects with the current batch for each sample.
            Each element of the Dataset is a tuple of size 2, specifying the
            data and label for BPTT respectively. Both items are of the same
            shape (seq_len, batch_size).
        """
        if self._last_batch == 'keep':
            coded = self._vocab[list(corpus)]
            sample_len = math.ceil(float(len(coded)) / self._batch_size)
            padding_size = _slice_pad_length(sample_len, self._seq_len + 1, 1) * \
                self._batch_size + sample_len * self._batch_size - len(coded)
            coded.extend([self._vocab[self._vocab.padding_token]] * int(padding_size))
            assert len(coded) % self._batch_size == 0
            assert not _slice_pad_length(len(coded) / self._batch_size, self._seq_len + 1, 1)
        else:
            sample_len = len(corpus) // self._batch_size
            coded = self._vocab[corpus[:sample_len * self._batch_size]]
        data = mx.nd.array(coded).reshape((self._batch_size, -1)).T
        batches = slice_sequence(data, self._seq_len + 1, overlap=1)

        return SimpleDataset(batches).transform(_split_data_label, lazy=False)


def _split_data_label(x):
    return x[:-1, :], x[1:, :]


class StreamBPTTBatchify:
    """Transform a Stream of CorpusDataset to BPTT batches.

    The corpus is transformed into batches of numericalized samples, in the way that the
    recurrent states from last batch connects with the current batch for each sample.

    Each sample is of shape `(seq_len, batch_size)`.

    For example, the following 4 sequences::

        a b c d <eos>
        e f g h i j <eos>
        k l m n <eos>
        o <eos>

    will generate 2 batches with seq_len = 5, batch_size = 2 as follow (transposed):

    batch_0.data.T::

        a b c d <eos>
        e f g h i

    batch_0.target.T::

        b c d <eos> k
        f g h i j

    batch_1.data.T::

        k l m n <eos>
        j <eos> o <eos> <padding>

    batch_1.target.T::

        l m n <eos> <padding>
        <eos> o <eos> <padding> <padding>

    Parameters
    ----------
    vocab : gluonnlp.Vocab
        The vocabulary to use for numericalizing the dataset. Each token will be mapped to the
        index according to the vocabulary.
    seq_len : int
        The length of each of the samples for truncated back-propagation-through-time (TBPTT).
    batch_size : int
        The number of samples in each batch.
    sampler : str, {'sequential', 'random'}, defaults to 'random'
        The sampler used to sample texts within a file.

        - 'sequential': SequentialSampler
        - 'random': RandomSampler
    last_batch : {'keep', 'discard'}
        How to handle the last batch if the remaining length is less than `seq_len`.

        - keep: A batch with less samples than previous batches is returned.
        - discard: The last batch is discarded if it's smaller than `(seq_len, batch_size)`.
    """

    def __init__(self,
                 vocab,
                 seq_len,
                 batch_size,
                 sampler='random',
                 last_batch='keep'):
        self._vocab = vocab
        self._seq_len = seq_len
        self._batch_size = batch_size
        self._sampler = sampler
        self._last_batch = last_batch
        if not self._vocab.padding_token:
            raise ValueError('Padding token must be specified in vocab for StreamBPTTBatchify.')

        if last_batch not in ['keep', 'discard']:
            raise ValueError(
                'Got invalid last_batch: "{}". Must be "keep" or "discard".'.
                format(last_batch))

    def _get_sampler(self, sampler):
        assert isinstance(
            sampler,
            str), 'Expected sampler to be a str, but got %s' % type(sampler)
        if sampler == 'random':
            return RandomSampler
        if sampler == 'sequential':
            return SequentialSampler
        raise ValueError(
            'sampler must be either "random" or "sequential", but got %s' %
            (sampler))

    def __call__(self, corpus):
        """Batchify a stream.

        Parameters
        ----------
        corpus : nlp.data.DatasetStream
            A stream of un-flattened CorpusDataset.

        Returns
        -------
        nlp.data.DataStream
            Batches of numericalized samples such that the recurrent states
            from last batch connects with the current batch for each sample.
            Each element of the Dataset is a tuple of data and label arrays for
            BPTT. They are of shape (seq_len, batch_size) respectively.
        """
        return _StreamBPTTBatchify(
            corpus, self._vocab, self._seq_len, self._batch_size,
            self._get_sampler(self._sampler), self._last_batch)


class _StreamBPTTBatchify(DataStream):
    def __init__(self, corpus, vocab, seq_len, batch_size, sampler,
                 last_batch):
        self._corpus = corpus
        self._vocab = vocab
        self._seq_len = seq_len
        self._batch_size = batch_size
        self._sampler = sampler
        self._last_batch = last_batch
        self._padding_idx = vocab[vocab.padding_token]

    def __iter__(self):
        def _init(data, target, value):
            """Init the data and target with values."""
            data[:] = value
            target[:] = value

        def _read(buffers, i, vocab, corpus):
            """Read a sentence from the corpus into i-th buffer."""
            if len(buffers[i]) <= 1:
                buffers[i].extend(vocab[next(corpus)])

        def _write(data, target, buffers, seq_len, i, length):
            """Write a sentence from i-th buffer to data and target."""
            num_tokens = len(buffers[i]) - 1
            num_tokens = min(num_tokens, seq_len - length)
            # fill in data and target
            data[i, length:length+num_tokens] = buffers[i][:num_tokens]
            target[i, length:length+num_tokens] = buffers[i][1:num_tokens+1]
            # trim sentence in the buffer if too long. Used for the next batch
            buffers[i] = buffers[i][num_tokens:]
            return num_tokens

        # stream states
        buffers = [[] for _ in range(self._batch_size)]
        has_next = True
        has_token_buffered = False
        data = np.empty([self._batch_size, self._seq_len], dtype=np.float32)
        target = np.empty([self._batch_size, self._seq_len], dtype=np.float32)
        corpus = itertools.chain.from_iterable(
            (corpus_dataset[idx] for idx in self._sampler(len(corpus_dataset)))
            for corpus_dataset in self._corpus)

        while has_next or has_token_buffered:
            _init(data, target, self._padding_idx)
            has_token_buffered = False
            for i in range(self._batch_size):
                length = 0
                try:
                    while length < self._seq_len:
                        _read(buffers, i, self._vocab, corpus)
                        num_tokens = _write(data, target, buffers, self._seq_len, i, length)
                        if len(buffers[i]) > 0:
                            has_token_buffered = True
                        length += num_tokens
                except StopIteration:
                    has_next = False
            if has_token_buffered or self._last_batch == 'keep':
                yield mx.nd.array(data).T, mx.nd.array(target).T



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Vocabulary class used in the original pre-trained ELMo models."""

# pylint: disable=consider-iterating-dictionary

__all__ = ['ELMoCharVocab']

class ELMoCharVocab:
    r"""ELMo special character vocabulary

    The vocab aims to map individual tokens to sequences of character ids, compatible with ELMo.
    To be consistent with previously trained models, we include it here.

    Specifically, char ids 0-255 come from utf-8 encoding bytes.
    Above 256 are reserved for special tokens.

    Parameters
    ----------
    bos_token : hashable object or None, default '<bos>'
        The representation for the special token of beginning-of-sequence token.
    eos_token : hashable object or None, default '<eos>'
        The representation for the special token of end-of-sequence token.

    Attributes
    ----------
    max_word_length : 50
        The maximum number of character a word contains is 50 in ELMo.
    bos_id : 256
        The index of beginning of the sentence character is 256 in ELMo.
    eos_id : 257
        The index of end of the sentence character is 257 in ELMo.
    bow_id : 258
        The index of beginning of the word character is 258 in ELMo.
    eow_id : 259
        The index of end of the word character is 259 in ELMo.
    pad_id : 260
        The index of padding character is 260 in ELMo.
    """
    max_word_length = 50
    max_word_chars = 48 # excluding bow and eow
    # char ids 0-255 come from utf-8 encoding bytes
    bos_id = 256
    eos_id = 257
    bow_id = 258
    eow_id = 259
    pad_id = 260

    def __init__(self, bos_token='<bos>', eos_token='<eos>'):
        self._bos_token = bos_token
        self._eos_token = eos_token
        self._id_dict = {bos_token: [ELMoCharVocab.bos_id],
                         eos_token: [ELMoCharVocab.eos_id]}

    def __getitem__(self, tokens):
        """Looks up indices of text tokens according to the vocabulary.

        Parameters
        ----------
        tokens : str or list of strs
            A source token or tokens to be converted.

        Returns
        -------
        int or list of ints
            A list of char indices or a list of list of char indices according to the vocabulary.
        """

        if not isinstance(tokens, (list, tuple)):
            return self._token_to_char_indices(tokens)
        else:
            return [self._token_to_char_indices(token) for token in tokens]

    def _token_to_char_indices(self, token):
        ids = [ELMoCharVocab.pad_id] * ELMoCharVocab.max_word_length
        ids[0] = ELMoCharVocab.bow_id
        word_ids = bytearray(token, 'utf-8', 'ignore')[:ELMoCharVocab.max_word_chars]
        word_ids = self._id_dict.get(token, word_ids)
        ids[1:(1+len(word_ids))] = word_ids
        ids[1+len(word_ids)] = ELMoCharVocab.eow_id
        return ids

    def __call__(self, tokens):
        """Looks up indices of text tokens according to the vocabulary.


        Parameters
        ----------
        tokens : str or list of strs
            A source token or tokens to be converted.


        Returns
        -------
        int or list of ints
            A list of char indices or a list of list of char indices according to the vocabulary.
        """

        return self[tokens]

    def __len__(self):
        return 262



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=
"""Machine translation datasets."""

__all__ = ['IWSLT2015', 'WMT2014', 'WMT2014BPE', 'WMT2016', 'WMT2016BPE']


import os
import zipfile
import shutil
import io

from mxnet.gluon.utils import download, check_sha1, _get_repo_file_url
from mxnet.gluon.data import ArrayDataset

from .dataset import TextLineDataset
from ..vocab import Vocab
from .registry import register
from ..base import get_home_dir


def _get_pair_key(src_lang, tgt_lang):
    return '_'.join(sorted([src_lang, tgt_lang]))


class _TranslationDataset(ArrayDataset):
    def __init__(self, namespace, segment, src_lang, tgt_lang, root):
        assert _get_pair_key(src_lang, tgt_lang) in self._archive_file, \
            'The given language combination: src_lang={}, tgt_lang={}, is not supported. ' \
            'Only supports language pairs = {}.'.format(
                src_lang, tgt_lang, str(self._archive_file.keys()))
        if isinstance(segment, str):
            assert segment in self._supported_segments, \
                'Only supports {} for the segment. Received segment={}'.format(
                    self._supported_segments, segment)
        else:
            for ele_segment in segment:
                assert ele_segment in self._supported_segments, \
                    'segment should only contain elements in {}. Received segment={}'.format(
                        self._supported_segments, segment)
        self._namespace = 'gluon/dataset/{}'.format(namespace)
        self._segment = segment
        self._src_lang = src_lang
        self._tgt_lang = tgt_lang
        self._src_vocab = None
        self._tgt_vocab = None
        self._pair_key = _get_pair_key(src_lang, tgt_lang)
        root = os.path.expanduser(root)
        if not os.path.isdir(root):
            os.makedirs(root)
        self._root = root
        if isinstance(segment, str):
            segment = [segment]
        src_corpus = []
        tgt_corpus = []
        for ele_segment in segment:
            [src_corpus_path, tgt_corpus_path] = self._get_data(ele_segment)
            src_corpus.extend(TextLineDataset(src_corpus_path))
            tgt_corpus.extend(TextLineDataset(tgt_corpus_path))
        # Filter 0-length src/tgt sentences
        src_lines = []
        tgt_lines = []
        for src_line, tgt_line in zip(list(src_corpus), list(tgt_corpus)):
            if len(src_line) > 0 and len(tgt_line) > 0:
                src_lines.append(src_line)
                tgt_lines.append(tgt_line)
        super(_TranslationDataset, self).__init__(src_lines, tgt_lines)

    def _fetch_data_path(self, file_name_hashs):
        archive_file_name, archive_hash = self._archive_file[self._pair_key]
        paths = []
        root = self._root
        for data_file_name, data_hash in file_name_hashs:
            path = os.path.join(root, data_file_name)
            if not os.path.exists(path) or not check_sha1(path, data_hash):
                downloaded_file_path = download(_get_repo_file_url(self._namespace,
                                                                   archive_file_name),
                                                path=root,
                                                sha1_hash=archive_hash)

                with zipfile.ZipFile(downloaded_file_path, 'r') as zf:
                    for member in zf.namelist():
                        filename = os.path.basename(member)
                        if filename:
                            dest = os.path.join(root, filename)
                            with zf.open(member) as source, \
                                    open(dest, 'wb') as target:
                                shutil.copyfileobj(source, target)
            paths.append(path)
        return paths

    def _get_data(self, segment):
        src_corpus_file_name, src_corpus_hash =\
            self._data_file[self._pair_key][segment + '_' + self._src_lang]
        tgt_corpus_file_name, tgt_corpus_hash =\
            self._data_file[self._pair_key][segment + '_' + self._tgt_lang]
        return self._fetch_data_path([(src_corpus_file_name, src_corpus_hash),
                                      (tgt_corpus_file_name, tgt_corpus_hash)])

    @property
    def src_vocab(self):
        """Source Vocabulary of the Dataset.

        Returns
        -------
        src_vocab : Vocab
            Source vocabulary.
        """
        if self._src_vocab is None:
            src_vocab_file_name, src_vocab_hash = \
                self._data_file[self._pair_key]['vocab' + '_' + self._src_lang]
            [src_vocab_path] = self._fetch_data_path([(src_vocab_file_name, src_vocab_hash)])
            with io.open(src_vocab_path, 'r', encoding='utf-8') as in_file:
                self._src_vocab = Vocab.from_json(in_file.read())
        return self._src_vocab

    @property
    def tgt_vocab(self):
        """Target Vocabulary of the Dataset.

        Returns
        -------
        tgt_vocab : Vocab
            Target vocabulary.
        """
        if self._tgt_vocab is None:
            tgt_vocab_file_name, tgt_vocab_hash = \
                self._data_file[self._pair_key]['vocab' + '_' + self._tgt_lang]
            [tgt_vocab_path] = self._fetch_data_path([(tgt_vocab_file_name, tgt_vocab_hash)])
            with io.open(tgt_vocab_path, 'r', encoding='utf-8') as in_file:
                self._tgt_vocab = Vocab.from_json(in_file.read())
        return self._tgt_vocab


@register(segment=['train', 'val', 'test'])
class IWSLT2015(_TranslationDataset):
    """Preprocessed IWSLT English-Vietnamese Translation Dataset.

    We use the preprocessed version provided in https://nlp.stanford.edu/projects/nmt/

    Parameters
    ----------
    segment : str or list of str, default 'train'
        Dataset segment. Options are 'train', 'val', 'test' or their combinations.
    src_lang : str, default 'en'
        The source language. Option for source and target languages are 'en' <-> 'vi'
    tgt_lang : str, default 'vi'
        The target language. Option for source and target languages are 'en' <-> 'vi'
    root : str, default '$MXNET_HOME/datasets/iwslt2015'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.
    """
    def __init__(self, segment='train', src_lang='en', tgt_lang='vi',
                 root=os.path.join(get_home_dir(), 'datasets', 'iwslt2015')):
        self._supported_segments = ['train', 'val', 'test']
        self._archive_file = {_get_pair_key('en', 'vi'):
                                  ('iwslt15.zip', '15a05df23caccb1db458fb3f9d156308b97a217b')}
        self._data_file = {_get_pair_key('en', 'vi'):
                               {'train_en': ('train.en',
                                             '675d16d057f2b6268fb294124b1646d311477325'),
                                'train_vi': ('train.vi',
                                             'bb6e21d4b02b286f2a570374b0bf22fb070589fd'),
                                'val_en': ('tst2012.en',
                                           'e381f782d637b8db827d7b4d8bb3494822ec935e'),
                                'val_vi': ('tst2012.vi',
                                           '4511988ce67591dc8bcdbb999314715f21e5a1e1'),
                                'test_en': ('tst2013.en',
                                            'd320db4c8127a85de81802f239a6e6b1af473c3d'),
                                'test_vi': ('tst2013.vi',
                                            'af212c48a68465ceada9263a049f2331f8af6290'),
                                'vocab_en': ('vocab.en.json',
                                             'b6f8e77a45f6dce648327409acd5d52b37a45d94'),
                                'vocab_vi' : ('vocab.vi.json',
                                              '9be11a9edd8219647754d04e0793d2d8c19dc852')}}
        super(IWSLT2015, self).__init__('iwslt2015', segment=segment, src_lang=src_lang,
                                        tgt_lang=tgt_lang, root=root)


@register(segment=['train', 'newstest2009', 'newstest2010', 'newstest2011', \
                   'newstest2012', 'newstest2013', 'newstest2014'])
class WMT2014(_TranslationDataset):
    """Translation Corpus of the WMT2014 Evaluation Campaign.

    http://www.statmt.org/wmt14/translation-task.html

    Parameters
    ----------
    segment : str or list of str, default 'train'
        Dataset segment. Options are 'train', 'newstest2009', 'newstest2010',
        'newstest2011', 'newstest2012', 'newstest2013', 'newstest2014' or their combinations
    src_lang : str, default 'en'
        The source language. Option for source and target languages are 'en' <-> 'de'
    tgt_lang : str, default 'de'
        The target language. Option for source and target languages are 'en' <-> 'de'
    full : bool, default False
        By default, we use the "filtered test sets" while if full is True, we use the "cleaned test
        sets".
    root : str, default '$MXNET_HOME/datasets/wmt2014'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.
    """
    def __init__(self, segment='train', src_lang='en', tgt_lang='de', full=False,
                 root=os.path.join(get_home_dir(), 'datasets', 'wmt2014')):
        self._supported_segments = ['train'] + ['newstest%d' % i for i in range(2009, 2015)]
        self._archive_file = {_get_pair_key('de', 'en'):
                                  ('wmt2014_de_en-b0e0e703.zip',
                                   'b0e0e7036217ffa94f4b35a5b5d2a96a27f680a4')}
        self._data_file = {_get_pair_key('de', 'en'):
                               {'train_en': ('train.en',
                                             'cec2d4c5035df2a54094076348eaf37e8b588a9b'),
                                'train_de': ('train.de',
                                             '6348764640ffc40992e7de89a8c48d32a8bcf458'),
                                'newstest2009_en': ('newstest2009.en',
                                                    'f8623af2de682924f9841488427e81c430e3ce60'),
                                'newstest2009_de': ('newstest2009.de',
                                                    'dec03f14cb47e726ccb19bec80c645d4a996f8a9'),
                                'newstest2010_en': ('newstest2010.en',
                                                    '5966eb13bd7cc8855cc6b40f9797607e36e9cc80'),
                                'newstest2010_de': ('newstest2010.de',
                                                    'b9af0cb004fa6996eda246d0173c191693b26025'),
                                'newstest2011_en': ('newstest2011.en',
                                                    '2c1d9d077fdbfe9d0e052a6e08a85ee7959479ab'),
                                'newstest2011_de': ('newstest2011.de',
                                                    'efbded3d175a9d472aa5938fe22afcc55c6055ff'),
                                'newstest2012_en': ('newstest2012.en',
                                                    '52f05ae725be45ee4012c6e208cef13614abacf1'),
                                'newstest2012_de': ('newstest2012.de',
                                                    'd9fe32143b88e6fe770843e15ee442a69ff6752d'),
                                'newstest2013_en': ('newstest2013.en',
                                                    '5dca5d02cf40278d8586ee7d58d58215253156a9'),
                                'newstest2013_de': ('newstest2013.de',
                                                    'ddda1e7b3270cb68108858640bfb619c37ede2ab'),
                                'newstest2014_en': ('newstest2014.src.en',
                                                    '610c5bb4cc866ad04ab1f6f80d740e1f4435027c'),
                                'newstest2014_de': ('newstest2014.ref.de',
                                                    '03b02c7f60c8509ba9bb4c85295358f7c9f00d2d')}}
        if full:
            self._data_file[_get_pair_key('de', 'en')]['newstest2014_en'] = \
                ('newstest2014.full.en', '528742a3a9690995d031f49d1dbb704844684976')
            self._data_file[_get_pair_key('de', 'en')]['newstest2014_de'] = \
                ('newstest2014.full.de', '2374b6a28cecbd965b73a9acc35a425e1ed81963')
        else:
            if src_lang == 'de':
                self._data_file[_get_pair_key('de', 'en')]['newstest2014_en'] = \
                    ('newstest2014.ref.en', 'cf23229ec6db8b85f240618d2a245f69afebed1f')
                self._data_file[_get_pair_key('de', 'en')]['newstest2014_de'] = \
                    ('newstest2014.src.de', '791d644b1a031268ca19600b2734a63c7bfcecc4')
        super(WMT2014, self).__init__('wmt2014', segment=segment, src_lang=src_lang,
                                      tgt_lang=tgt_lang,
                                      root=os.path.join(root, _get_pair_key(src_lang, tgt_lang)))


@register(segment=['train', 'newstest2009', 'newstest2010', 'newstest2011', \
                   'newstest2012', 'newstest2013', 'newstest2014'])
class WMT2014BPE(_TranslationDataset):
    """Preprocessed Translation Corpus of the WMT2014 Evaluation Campaign.

    We preprocess the dataset by adapting
    https://github.com/tensorflow/nmt/blob/master/nmt/scripts/wmt16_en_de.sh

    Parameters
    ----------
    segment : str or list of str, default 'train'
        Dataset segment. Options are 'train', 'newstest2009', 'newstest2010',
        'newstest2011', 'newstest2012', 'newstest2013', 'newstest2014' or their combinations
    src_lang : str, default 'en'
        The source language. Option for source and target languages are 'en' <-> 'de'
    tgt_lang : str, default 'de'
        The target language. Option for source and target languages are 'en' <-> 'de'
    full : bool, default False
        In default, we use the test dataset in http://statmt.org/wmt14/test-filtered.tgz.
        When full is True, we use the test dataset in http://statmt.org/wmt14/test-full.tgz
    root : str, default '$MXNET_HOME/datasets/wmt2014'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.
    """
    def __init__(self, segment='train', src_lang='en', tgt_lang='de', full=False,
                 root=os.path.join(get_home_dir(), 'datasets', 'wmt2014')):
        self._supported_segments = ['train'] + ['newstest%d' % i for i in range(2009, 2015)]
        self._archive_file = {_get_pair_key('de', 'en'):
                                  ('wmt2014bpe_de_en-ace8f41c.zip',
                                   'ace8f41c22c0da8729ff15f40d416ebd16738979')}
        self._data_file = {_get_pair_key('de', 'en'):
                               {'train_en': ('train.tok.clean.bpe.32000.en',
                                             'e3f093b64468db7084035c9650d9eecb86a3db5f'),
                                'train_de': ('train.tok.clean.bpe.32000.de',
                                             '60703ad088706a3d9d1f3328889c6f4725a36cfb'),
                                'newstest2009_en': ('newstest2009.tok.bpe.32000.en',
                                                    '5678547f579528a8716298e895f886e3976085e1'),
                                'newstest2009_de': ('newstest2009.tok.bpe.32000.de',
                                                    '32caa69023eac1750a0036780f9d511d979aed2c'),
                                'newstest2010_en': ('newstest2010.tok.bpe.32000.en',
                                                    '813103f7b4b472cf213fe3b2c3439e267dbc4afb'),
                                'newstest2010_de': ('newstest2010.tok.bpe.32000.de',
                                                    '972076a897ecbc7a3acb639961241b33fd58a374'),
                                'newstest2011_en': ('newstest2011.tok.bpe.32000.en',
                                                    'c3de2d72d5e7bdbe848839c55c284fece90464ce'),
                                'newstest2011_de': ('newstest2011.tok.bpe.32000.de',
                                                    '7a8722aeedacd99f1aa8dffb6d8d072430048011'),
                                'newstest2012_en': ('newstest2012.tok.bpe.32000.en',
                                                    '876ad3c72e33d8e1ed14f5362f97c771ce6a9c7f'),
                                'newstest2012_de': ('newstest2012.tok.bpe.32000.de',
                                                    '57467fcba8442164d058a05eaf642a1da1d92c13'),
                                'newstest2013_en': ('newstest2013.tok.bpe.32000.en',
                                                    'de06a155c3224674b2434f3ff3b2c4a4a293d238'),
                                'newstest2013_de': ('newstest2013.tok.bpe.32000.de',
                                                    '094084989128dd091a2fe2a5818a86bc99ecc0e7'),
                                'newstest2014_en': ('newstest2014.tok.bpe.32000.src.en',
                                                    '347cf4d3d5c3c46ca1220247d22c07aa90092bd9'),
                                'newstest2014_de': ('newstest2014.tok.bpe.32000.ref.de',
                                                    'f66b80a0c460c524ec42731e527c54aab5507a66'),
                                'vocab_en': ('vocab.bpe.32000.json',
                                             '71413f497ce3a0fa691c55277f367e5d672b27ee'),
                                'vocab_de': ('vocab.bpe.32000.json',
                                             '71413f497ce3a0fa691c55277f367e5d672b27ee')}}
        if full:
            self._data_file[_get_pair_key('de', 'en')]['newstest2014_en'] = \
                ('newstest2014.tok.bpe.32000.full.en', '6c398b61641cd39f186b417c54b171876563193f')
            self._data_file[_get_pair_key('de', 'en')]['newstest2014_de'] = \
                ('newstest2014.tok.bpe.32000.full.de', 'b890a8dfc2146dde570fcbcb42e4157292e95251')
        else:
            if src_lang == 'de':
                self._data_file[_get_pair_key('de', 'en')]['newstest2014_en'] = \
                    ('newstest2014.tok.bpe.32000.ref.en',
                     'cd416085db722bf07cbba4ff29942fe94e966023')
                self._data_file[_get_pair_key('de', 'en')]['newstest2014_de'] = \
                    ('newstest2014.tok.bpe.32000.src.de',
                     '9274d31f92141933f29a405753d5fae051fa5725')
        super(WMT2014BPE, self).__init__('wmt2014', segment=segment, src_lang=src_lang,
                                         tgt_lang=tgt_lang,
                                         root=os.path.join(root, _get_pair_key(src_lang, tgt_lang)))


@register(segment=['train', 'newstest2012', 'newstest2013', 'newstest2014', \
                   'newstest2015', 'newstest2016'])
class WMT2016(_TranslationDataset):
    """Translation Corpus of the WMT2016 Evaluation Campaign.

    Parameters
    ----------
    segment : str or list of str, default 'train'
        Dataset segment. Options are 'train', 'newstest2012', 'newstest2013',
        'newstest2014', 'newstest2015', 'newstest2016' or their combinations
    src_lang : str, default 'en'
        The source language. Option for source and target languages are 'en' <-> 'de'
    tgt_lang : str, default 'de'
        The target language. Option for source and target languages are 'en' <-> 'de'
    root : str, default '$MXNET_HOME/datasets/wmt2016'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.
    """
    def __init__(self, segment='train', src_lang='en', tgt_lang='de',
                 root=os.path.join(get_home_dir(), 'datasets', 'wmt2016')):
        self._supported_segments = ['train'] + ['newstest%d' % i for i in range(2012, 2017)]
        self._archive_file = {_get_pair_key('de', 'en'):
                                  ('wmt2016_de_en-88767407.zip',
                                   '887674077b951ce949fe3e597086b826bd7574d8')}
        self._data_file = {_get_pair_key('de', 'en'):
                               {'train_en': ('train.en',
                                             '1be6d00c255c57183305276c5de60771e201d3b0'),
                                'train_de': ('train.de',
                                             '4eec608b8486bfb65b61bda237b0c9b3c0f66f17'),
                                'newstest2012_en': ('newstest2012.en',
                                                    '52f05ae725be45ee4012c6e208cef13614abacf1'),
                                'newstest2012_de': ('newstest2012.de',
                                                    'd9fe32143b88e6fe770843e15ee442a69ff6752d'),
                                'newstest2013_en': ('newstest2013.en',
                                                    '5dca5d02cf40278d8586ee7d58d58215253156a9'),
                                'newstest2013_de': ('newstest2013.de',
                                                    'ddda1e7b3270cb68108858640bfb619c37ede2ab'),
                                'newstest2014_en': ('newstest2014.en',
                                                    '528742a3a9690995d031f49d1dbb704844684976'),
                                'newstest2014_de': ('newstest2014.de',
                                                    '2374b6a28cecbd965b73a9acc35a425e1ed81963'),
                                'newstest2015_en': ('newstest2015.en',
                                                    'bf90439b209a496128995c4b948ad757979d0756'),
                                'newstest2015_de': ('newstest2015.de',
                                                    'd69ac825fe3d5796b4990b969ad71903a38a0cd1'),
                                'newstest2016_en': ('newstest2016.en',
                                                    'a99c145d5214eb1645b56d21b02a541fbe7eb3c2'),
                                'newstest2016_de': ('newstest2016.de',
                                                    'fcdd3104f21eb4b9c49ba8ddef46d9b2d472b3fe')}}
        super(WMT2016, self).__init__('wmt2016', segment=segment, src_lang=src_lang,
                                      tgt_lang=tgt_lang,
                                      root=os.path.join(root, _get_pair_key(src_lang, tgt_lang)))


@register(segment=['train', 'newstest2012', 'newstest2013', 'newstest2014', \
                   'newstest2015', 'newstest2016'])
class WMT2016BPE(_TranslationDataset):
    """Preprocessed Translation Corpus of the WMT2016 Evaluation Campaign.

    We use the preprocessing script in
    https://github.com/tensorflow/nmt/blob/master/nmt/scripts/wmt16_en_de.sh

    Parameters
    ----------
    segment : str or list of str, default 'train'
        Dataset segment. Options are 'train', 'newstest2012', 'newstest2013',
        'newstest2014', 'newstest2015', 'newstest2016' or their combinations
    src_lang : str, default 'en'
        The source language. Option for source and target languages are 'en' <-> 'de'
    tgt_lang : str, default 'de'
        The target language. Option for source and target languages are 'en' <-> 'de'
    root : str, default '$MXNET_HOME/datasets/wmt2016'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.
    """
    def __init__(self, segment='train', src_lang='en', tgt_lang='de',
                 root=os.path.join(get_home_dir(), 'datasets', 'wmt2016')):
        self._supported_segments = ['train'] + ['newstest%d' % i for i in range(2012, 2017)]
        self._archive_file = {_get_pair_key('de', 'en'):
                                  ('wmt2016bpe_de_en-8cf0dbf6.zip',
                                   '8cf0dbf6a102381443a472bcf9f181299231b496')}
        self._data_file = {_get_pair_key('de', 'en'):
                               {'train_en': ('train.tok.clean.bpe.32000.en',
                                             '56f37cb4d68c2f83efd6a0c555275d1fe09f36b5'),
                                'train_de': ('train.tok.clean.bpe.32000.de',
                                             '58f30a0ba7f80a8840a5cf3deff3c147de7d3f68'),
                                'newstest2012_en': ('newstest2012.tok.bpe.32000.en',
                                                    '25ed9ad228a236f57f97bf81db1bb004bedb7f33'),
                                'newstest2012_de': ('newstest2012.tok.bpe.32000.de',
                                                    'bb5622831ceea1894966fa993ebcd882cc461943'),
                                'newstest2013_en': ('newstest2013.tok.bpe.32000.en',
                                                    'fa03fe189fe68cb25014c5e64096ac8daf2919fa'),
                                'newstest2013_de': ('newstest2013.tok.bpe.32000.de',
                                                    '7d10a884499d352c2fea6f1badafb40473737640'),
                                'newstest2014_en': ('newstest2014.tok.bpe.32000.en',
                                                    '7b8ea824021cc5291e6a54bb32a1fc27c2955588'),
                                'newstest2014_de': ('newstest2014.tok.bpe.32000.de',
                                                    'd84497d4c425fa4e9b2b6be4b62c763086410aad'),
                                'newstest2015_en': ('newstest2015.tok.bpe.32000.en',
                                                    'ca335076f67b2f9b98848f8abc2cd424386f2309'),
                                'newstest2015_de': ('newstest2015.tok.bpe.32000.de',
                                                    'e633a3fb74506eb498fcad654d82c9b1a0a347b3'),
                                'newstest2016_en': ('newstest2016.tok.bpe.32000.en',
                                                    '5a5e36a6285823035b642aef7c1a9ec218da59f7'),
                                'newstest2016_de': ('newstest2016.tok.bpe.32000.de',
                                                    '135a79acb6a4f8fad0cbf5f74a15d9c0b5bf8c73'),
                                'vocab_en': ('vocab.bpe.32000.json',
                                             '1c5aea0a77cad592c4e9c1136ec3b70ceeff4e8c'),
                                'vocab_de': ('vocab.bpe.32000.json',
                                             '1c5aea0a77cad592c4e9c1136ec3b70ceeff4e8c')}}
        super(WMT2016BPE, self).__init__('wmt2016', segment=segment, src_lang=src_lang,
                                         tgt_lang=tgt_lang,
                                         root=os.path.join(root, _get_pair_key(src_lang, tgt_lang)))



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Candidate samplers"""

__all__ = ['UnigramCandidateSampler']

import mxnet as mx
import numpy as np


class UnigramCandidateSampler(mx.gluon.HybridBlock):
    """Unigram Candidate Sampler

    Draw random samples from a unigram distribution with specified weights
    using the alias method.

    Parameters
    ----------
    weights : mx.nd.NDArray
        Unnormalized class probabilities. Samples are drawn and returned on the
        same context as weights.context.
    dtype : str or np.dtype, default 'float32'
        Data type of the candidates. Make sure that the dtype precision is
        large enough to represent the size of your weights array precisely. For
        example, float32 can not distinguish 2**24 from 2**24 + 1.

    """

    def __init__(self, weights, dtype='float32'):
        super(UnigramCandidateSampler, self).__init__()
        self._dtype = dtype
        self.N = weights.size

        if (np.dtype(dtype) == np.float32 and weights.size > 2**24) or \
           (np.dtype(dtype) == np.float16 and weights.size > 2**11):
            s = 'dtype={dtype} can not represent all weights'
            raise ValueError(s.format(dtype=dtype))

        total_weights = weights.sum()
        prob = (weights * self.N / total_weights).asnumpy().tolist()
        alias = [0] * self.N

        # sort the data into the outcomes with probabilities
        # that are high and low than 1/N.
        low = []
        high = []
        for i in range(self.N):
            if prob[i] < 1.0:
                low.append(i)
            else:
                high.append(i)

        # pair low with high
        while len(low) > 0 and len(high) > 0:
            l = low.pop()
            h = high.pop()

            alias[l] = h
            prob[h] = prob[h] - (1.0 - prob[l])

            if prob[h] < 1.0:
                low.append(h)
            else:
                high.append(h)

        for i in low + high:
            prob[i] = 1
            alias[i] = i

        # store
        prob = mx.nd.array(prob, dtype='float64')
        alias = mx.nd.array(alias, dtype='float64')
        self.prob = self.params.get_constant('prob', prob)
        self.alias = self.params.get_constant('alias', alias)

    def __repr__(self):
        s = '{block_name}({len_weights}, {dtype})'
        return s.format(block_name=self.__class__.__name__, len_weights=self.N,
                        dtype=self._dtype)

    # pylint: disable=arguments-differ, unused-argument
    def hybrid_forward(self, F, candidates_like, prob, alias):
        """Draw samples from uniform distribution and return sampled candidates.

        Parameters
        ----------
        candidates_like: mxnet.nd.NDArray or mxnet.sym.Symbol
            This input specifies the shape of the to be sampled candidates. #

        Returns
        -------
        samples: mxnet.nd.NDArray or mxnet.sym.Symbol
            The sampled candidates of shape candidates_like.shape. Candidates
            are sampled based on the weights specified on creation of the
            UnigramCandidateSampler.
        """
        candidates_flat = candidates_like.reshape((-1, )).astype('float64')
        idx = F.random.uniform_like(candidates_flat, low=0, high=self.N).floor()
        prob = F.gather_nd(prob, idx.reshape((1, -1)))
        alias = F.gather_nd(alias, idx.reshape((1, -1)))
        where = F.random.uniform_like(candidates_flat) < prob
        hit = idx * where
        alt = alias * (1 - where)
        candidates = (hit + alt).reshape_like(candidates_like)

        return candidates.astype(self._dtype)



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Building blocks and utility for models."""
__all__ = ['RNNCellLayer', 'L2Normalization', 'GELU']

import math
from mxnet import ndarray
from mxnet.gluon import Block, HybridBlock


class RNNCellLayer(Block):
    """A block that takes an rnn cell and makes it act like rnn layer.

    Parameters
    ----------
    rnn_cell : Cell
        The cell to wrap into a layer-like block.
    layout : str, default 'TNC'
        The output layout of the layer.
    """
    def __init__(self, rnn_cell, layout='TNC', **kwargs):
        super(RNNCellLayer, self).__init__(**kwargs)
        self.cell = rnn_cell
        assert layout in ('TNC', 'NTC'), \
            'Invalid layout %s; must be one of ["TNC" or "NTC"]'%layout
        self._layout = layout
        self._axis = layout.find('T')
        self._batch_axis = layout.find('N')

    def forward(self, inputs, states=None): # pylint: disable=arguments-differ
        """Defines the forward computation. Arguments can be either
        :py:class:`NDArray` or :py:class:`Symbol`."""
        batch_size = inputs.shape[self._batch_axis]
        skip_states = states is None
        if skip_states:
            states = self.cell.begin_state(batch_size, ctx=inputs.context)
        if isinstance(states, ndarray.NDArray):
            states = [states]
        for state, info in zip(states, self.cell.state_info(batch_size)):
            if state.shape != info['shape']:
                raise ValueError(
                    'Invalid recurrent state shape. Expecting %s, got %s.'%(
                        str(info['shape']), str(state.shape)))
        states = sum(zip(*((j for j in i) for i in states)), ())
        outputs, states = self.cell.unroll(
            inputs.shape[self._axis], inputs, states,
            layout=self._layout, merge_outputs=True)

        if skip_states:
            return outputs
        return outputs, states


class L2Normalization(HybridBlock):
    """Normalize the input array by dividing the L2 norm along the given axis.

    ..code

        out = data / (sqrt(sum(data**2, axis)) + eps)

    Parameters
    ----------
    axis : int, default -1
        The axis to compute the norm value.
    eps : float, default 1E-6
        The epsilon value to avoid dividing zero
    """
    def __init__(self, axis=-1, eps=1E-6, **kwargs):
        super(L2Normalization, self).__init__(**kwargs)
        self._axis = axis
        self._eps = eps

    def hybrid_forward(self, F, x):  # pylint: disable=arguments-differ
        ret = F.broadcast_div(x, F.norm(x, axis=self._axis, keepdims=True) + self._eps)
        return ret


class GELU(HybridBlock):
    """Gaussian Error Linear Unit.

    This is a smoother version of the RELU.
    https://arxiv.org/abs/1606.08415

    Parameters
    ----------
    approximate : bool, default False
        If True, use tanh approximation to calculate gelu. If False, use erf.

    """

    def __init__(self, approximate=False, prefix=None, params=None):
        super().__init__(prefix=prefix, params=params)
        self._approximate = approximate

    def hybrid_forward(self, F, x):  # pylint: disable=arguments-differ
        """

        Parameters
        ----------
        Inputs:
            - **data**: input tensor with arbitrary shape.
        Outputs:
            - **out**: output tensor with the same shape as `data`.
        """
        if not self._approximate:
            return F.LeakyReLU(x, act_type='gelu')
        else:
            return 0.5 * x * (1 + F.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * (x ** 3))))

    def __repr__(self):
        s = '{name}()'
        return s.format(name=self.__class__.__name__)




# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=
"""Sentiment analysis datasets."""

__all__ = ['IMDB', 'MR', 'TREC', 'SUBJ', 'SST_1', 'SST_2', 'CR', 'MPQA']

import json
import os
import shutil
import zipfile

from mxnet.gluon.data import SimpleDataset
from mxnet.gluon.utils import download, check_sha1, _get_repo_file_url
from .registry import register
from ..base import get_home_dir

class SentimentDataset(SimpleDataset):
    """Base class for sentiment analysis data sets.

    Parameters
    ----------
    segment : str
        Dataset segment.
    root : str
        Path to temp folder for storing data.
    """
    def __init__(self, segment, root):
        root = os.path.expanduser(root)
        if not os.path.isdir(root):
            os.makedirs(root)
        self._root = root
        self._segment = segment
        self._get_data()
        super(SentimentDataset, self).__init__(self._read_data())

    def _get_data(self):
        """Load data from the file. Do nothing if data was loaded before.
        """
        (data_archive_name, archive_hash), (data_name, data_hash) \
            = self._data_file()[self._segment]
        data_path = os.path.join(self._root, data_name)

        if not os.path.exists(data_path) or not check_sha1(data_path, data_hash):
            file_path = download(_get_repo_file_url(self._repo_dir(), data_archive_name),
                                 path=self._root, sha1_hash=archive_hash)

            with zipfile.ZipFile(file_path, 'r') as zf:
                for member in zf.namelist():
                    filename = os.path.basename(member)
                    if filename:
                        dest = os.path.join(self._root, filename)
                        with zf.open(member) as source, open(dest, 'wb') as target:
                            shutil.copyfileobj(source, target)

    def _read_data(self):
        (_, _), (data_file_name, _) = self._data_file()[self._segment]

        with open(os.path.join(self._root, data_file_name)) as f:
            samples = json.load(f)
        return samples

    def _data_file(self):
        raise NotImplementedError

    def _repo_dir(self):
        raise NotImplementedError


@register(segment=['train', 'test', 'unsup'])
class IMDB(SimpleDataset):
    """IMDB reviews for sentiment analysis.

    From
    http://ai.stanford.edu/~amaas/data/sentiment/

    Positive classes have label values in [7, 10]. Negative classes have label values in [1, 4].
    All samples in unsupervised set have labels with value 0.

    Parameters
    ----------
    segment : str, default 'train'
        Dataset segment. Options are 'train', 'test', and 'unsup' for unsupervised.
    root : str, default '$MXNET_HOME/datasets/imdb'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> imdb = gluonnlp.data.IMDB('test', root='./datasets/imdb')
    -etc-
    >>> len(imdb)
    25000
    >>> len(imdb[0])
    2
    >>> type(imdb[0][0]), type(imdb[0][1])
    (<class 'str'>, <class 'int'>)
    >>> imdb[0][0][:75]
    'I went and saw this movie last night after being coaxed to by a few friends'
    >>> imdb[0][1]
    10
    >>> imdb = gluonnlp.data.IMDB('unsup', root='./datasets/imdb')
    -etc-
    >>> len(imdb)
    50000
    >>> len(imdb[0])
    2
    >>> type(imdb[0][0]), type(imdb[0][1])
    (<class 'str'>, <class 'int'>)
    >>> imdb[0][0][:70]
    'I admit, the great majority of films released before say 1933 are just'
    >>> imdb[0][1]
    0
    """
    def __init__(self, segment='train', root=os.path.join(get_home_dir(), 'datasets', 'imdb')):
        self._data_file = {'train': ('train.json',
                                     '516a0ba06bca4e32ee11da2e129f4f871dff85dc'),
                           'test': ('test.json',
                                    '7d59bd8899841afdc1c75242815260467495b64a'),
                           'unsup': ('unsup.json',
                                     'f908a632b7e7d7ecf113f74c968ef03fadfc3c6c')}
        root = os.path.expanduser(root)
        if not os.path.isdir(root):
            os.makedirs(root)
        self._root = root
        self._segment = segment
        self._get_data()
        super(IMDB, self).__init__(self._read_data())

    def _get_data(self):
        data_file_name, data_hash = self._data_file[self._segment]
        root = self._root
        path = os.path.join(root, data_file_name)
        if not os.path.exists(path) or not check_sha1(path, data_hash):
            download(_get_repo_file_url('gluon/dataset/imdb', data_file_name),
                     path=root, sha1_hash=data_hash)

    def _read_data(self):
        with open(os.path.join(self._root, self._segment+'.json')) as f:
            samples = json.load(f)
        return samples


@register()
class MR(SentimentDataset):
    """Movie reviews for sentiment analysis.

    From
    https://www.cs.cornell.edu/people/pabo/movie-review-data/

    Positive class has label value 1. Negative class has label value 0.

    Parameters
    ----------
    root : str, default '$MXNET_HOME/datasets/mr'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> mr = gluonnlp.data.MR(root='./datasets/mr')
    -etc-
    >>> len(mr)
    10662
    >>> len(mr[3])
    2
    >>> type(mr[3][0]), type(mr[3][1])
    (<class 'str'>, <class 'int'>)
    >>> mr[3][0][:55]
    'if you sometimes like to go to the movies to have fun ,'
    >>> mr[3][1]
    1
    """
    def __init__(self, root=os.path.join(get_home_dir(), 'datasets', 'mr')):
        super(MR, self).__init__('all', root)

    def _data_file(self):
        return {'all': (('all-7606efec.zip', '0fcbaffe0bac94733e6497f700196585f03fa89e'),
                        ('all-7606efec.json', '7606efec578d9613f5c38bf2cef8d3e4e6575b2c '))}

    def _repo_dir(self):
        return 'gluon/dataset/mr'


@register(segment=['train', 'test'])
class TREC(SentimentDataset):
    """Question dataset for question classification.

    From
    http://cogcomp.cs.illinois.edu/Data/QA/QC/

    Class labels are (http://cogcomp.org/Data/QA/QC/definition.html):
        - DESCRIPTION: 0
        - ENTITY: 1
        - ABBREVIATION: 2
        - HUMAN: 3
        - LOCATION: 4
        - NUMERIC: 5

    The first space-separated token in the text of each sample is the fine-grain label.

    Parameters
    ----------
    segment : str, default 'train'
        Dataset segment. Options are 'train' and 'test'.
    root : str, default '$MXNET_HOME/datasets/trec'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> trec = gluonnlp.data.TREC('test', root='./datasets/trec')
    -etc-
    >>> len(trec)
    500
    >>> len(trec[0])
    2
    >>> type(trec[0][0]), type(trec[0][1])
    (<class 'str'>, <class 'int'>)
    >>> trec[0][0]
    'How far is it from Denver to Aspen ?'
    >>> (trec[0][1], trec[0][0].split()[0])
    (5, 'How')
    """
    def __init__(self, segment='train', root=os.path.join(get_home_dir(), 'datasets', 'trec')):
        super(TREC, self).__init__(segment, root)

    def _data_file(self):
        return {'train': (('train-1776132f.zip', '337d3f43a56ec26f5773c6fc406ef19fb4cd3c92'),
                          ('train-1776132f.json', '1776132fb2fc0ed2dc91b62f7817a4e071a3c7de')),
                'test': (('test-ff9ad0ce.zip', '57f03aaee2651ca05f1f9fc5731ba7e9ad98e38a'),
                         ('test-ff9ad0ce.json', 'ff9ad0ceb44d8904663fee561804a8dd0edc1b15'))}

    def _repo_dir(self):
        return 'gluon/dataset/trec'


@register()
class SUBJ(SentimentDataset):
    """Subjectivity dataset for sentiment analysis.

    Positive class has label value 1. Negative class has label value 0.

    Parameters
    ----------
    root : str, default '$MXNET_HOME/datasets/subj'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> subj = gluonnlp.data.SUBJ(root='./datasets/subj')
    -etc-
    >>> len(subj)
    10000
    >>> len(subj[0])
    2
    >>> type(subj[0][0]), type(subj[0][1])
    (<class 'str'>, <class 'int'>)
    >>> subj[0][0][:60]
    'its impressive images of crematorium chimney fires and stack'
    >>> subj[0][1]
    1
    """
    def __init__(self, root=os.path.join(get_home_dir(), 'datasets', 'subj')):
        super(SUBJ, self).__init__('all', root)

    def _data_file(self):
        return {'all': (('all-9e7bd1da.zip', '8b0d95c2fc885cc38e4ad776d7429183f3ef632b'),
                        ('all-9e7bd1da.json', '9e7bd1daa359c24abe1fac767d0e0af7bc114045'))}

    def _repo_dir(self):
        return 'gluon/dataset/subj'


@register(segment=['train', 'dev', 'test'])
class SST_1(SentimentDataset):
    """Stanford Sentiment Treebank: an extension of the MR data set.
    However, train/dev/test splits are provided and labels are fine-grained
    (very positive, positive, neutral, negative, very negative).

    From
    http://nlp.stanford.edu/sentiment/

    Class labels are:
        - very positive: 4
        - positive: 3
        - neutral: 2
        - negative: 1
        - very negative: 0

    Parameters
    ----------
    segment : str, default 'train'
        Dataset segment. Options are 'train' and 'test'.
    root : str, default '$MXNET_HOME/datasets/sst-1'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> sst_1 = gluonnlp.data.SST_1('test', root='./datasets/sst_1')
    -etc-
    >>> len(sst_1)
    2210
    >>> len(sst_1[0])
    2
    >>> type(sst_1[0][0]), type(sst_1[0][1])
    (<class 'str'>, <class 'int'>)
    >>> sst_1[0][0][:73]
    'no movement , no yuks , not much of anything .'
    >>> sst_1[0][1]
    1
    """
    def __init__(self, segment='train', root=os.path.join(get_home_dir(), 'datasets', 'sst-1')):
        super(SST_1, self).__init__(segment, root)

    def _data_file(self):
        return {'train': (('train-638f9352.zip', '0a039010449772700c0e270c7095362403dc486a'),
                          ('train-638f9352.json', '638f935251c0474e93d4aa50fda0c900faf02bba')),
                'dev': (('dev-820ac954.zip', 'e4b7899ef5d37a6bf01d8ec1115ba20b8419b96f'),
                        ('dev-820ac954.json', '820ac954b14b4f7d947e25f7a99249618d7962ee')),
                'test': (('test-ab593ae9.zip', 'd3736db56cdc7293c38435557697c2407652525d'),
                         ('test-ab593ae9.json', 'ab593ae9628f94af4f698654158ded1488b1de3b'))}

    def _repo_dir(self):
        return 'gluon/dataset/sst-1'


@register(segment=['train', 'dev', 'test'])
class SST_2(SentimentDataset):
    """Stanford Sentiment Treebank: an extension of the MR data set.
    Same as the SST-1 data set except that neutral reviews are removed
    and labels are binary (positive, negative).

    From
    http://nlp.stanford.edu/sentiment/

    Positive class has label value 1. Negative class has label value 0.

    Parameters
    ----------
    segment : str, default 'train'
        Dataset segment. Options are 'train' and 'test'.
    root : str, default '$MXNET_HOME/datasets/sst-2'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> sst_2 = gluonnlp.data.SST_2('test', root='./datasets/sst_2')
    -etc-
    >>> len(sst_2)
    1821
    >>> len(sst_2[0])
    2
    >>> type(sst_2[0][0]), type(sst_2[0][1])
    (<class 'str'>, <class 'int'>)
    >>> sst_2[0][0][:65]
    'no movement , no yuks , not much of anything .'
    >>> sst_2[0][1]
    0
    """
    def __init__(self, segment='train', root=os.path.join(get_home_dir(), 'datasets', 'sst-2')):
        super(SST_2, self).__init__(segment, root)

    def _data_file(self):
        return {'train': (('train-61f1f238.zip', 'f27a9ac6a7c9208fb7f024b45554da95639786b3'),
                          ('train-61f1f238.json', '61f1f23888652e11fb683ac548ed0be8a87dddb1')),
                'dev': (('dev-65511587.zip', '8c74911f0246bd88dc0ced2619f95f10db09dc98'),
                        ('dev-65511587.json', '655115875d83387b61f9701498143724147a1fc9')),
                'test': (('test-a39c1db6.zip', '4b7f1648207ec5dffb4e4783cf1f48d6f36ba4db'),
                         ('test-a39c1db6.json', 'a39c1db6ecc3be20bf2563bf2440c3c06887a2df'))}

    def _repo_dir(self):
        return 'gluon/dataset/sst-2'

@register()
class CR(SentimentDataset):
    """
    Customer reviews of various products (cameras, MP3s etc.). The task is to
    predict positive/negative reviews.

    Positive class has label value 1. Negative class has label value 0.

    Parameters
    ----------
    root : str, default '$MXNET_HOME/datasets/cr'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> cr = gluonnlp.data.CR(root='./datasets/cr')
    -etc-
    >>> len(cr)
    3775
    >>> len(cr[3])
    2
    >>> type(cr[3][0]), type(cr[3][1])
    (<class 'str'>, <class 'int'>)
    >>> cr[3][0][:55]
    'i know the saying is " you get what you pay for " but a'
    >>> cr[3][1]
    0
    """
    def __init__(self, root=os.path.join(get_home_dir(), 'datasets', 'cr')):
        super(CR, self).__init__('all', root)

    def _data_file(self):
        return {'all': (('all-0c9633c6.zip', 'c662e2f9115d74e1fcc7c896fa3e2dc5ee7688e7'),
                        ('all-0c9633c6.json', '0c9633c695d29b18730eddff965c850425996edf'))}

    def _repo_dir(self):
        return 'gluon/dataset/cr'

@register()
class MPQA(SentimentDataset):
    """
    Opinion polarity detection subtask of the MPQA dataset.

    From
    http://www.cs.pitt.edu/mpqa/

    Positive class has label value 1. Negative class has label value 0.

    Parameters
    ----------
    root : str, default '$MXNET_HOME/datasets/mpqa'
        Path to temp folder for storing data.
        MXNET_HOME defaults to '~/.mxnet'.

    Examples
    --------
    >>> mpqa = gluonnlp.data.MPQA(root='./datasets/mpqa')
    -etc-
    >>> len(mpqa)
    10606
    >>> len(mpqa[3])
    2
    >>> type(mpqa[3][0]), type(mpqa[3][1])
    (<class 'str'>, <class 'int'>)
    >>> mpqa[3][0][:25]
    'many years of decay'
    >>> mpqa[3][1]
    0
    """
    def __init__(self, root=os.path.join(get_home_dir(), 'datasets', 'mpqa')):
        super(MPQA, self).__init__('all', root)

    def _data_file(self):
        return {'all': (('all-bcbfeed8.zip', 'e07ae226cfe4713328eeb9660b261b9852ff5865'),
                        ('all-bcbfeed8.json', 'bcbfeed8b8767a564bdc428486ef18c1ba4dc536'))}

    def _repo_dir(self):
        return 'gluon/dataset/mpqa'



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""A registry for datasets

The registry makes it simple to construct a dataset given its name.

"""
__all__ = ['register', 'create', 'list_datasets']

import inspect

from mxnet import registry
from mxnet.gluon.data import Dataset

_REGSITRY_NAME_KWARGS = {}


def register(class_=None, **kwargs):
    """Registers a dataset with segment specific hyperparameters.

    When passing keyword arguments to `register`, they are checked to be valid
    keyword arguments for the registered Dataset class constructor and are
    saved in the registry. Registered keyword arguments can be retrieved with
    the `list_datasets` function.

    All arguments that result in creation of separate datasets should be
    registered. Examples are datasets divided in different segments or
    categories, or datasets containing multiple languages.

    Once registered, an instance can be created by calling
    :func:`~gluonnlp.data.create` with the class name.

    Parameters
    ----------
    **kwargs : list or tuple of allowed argument values
        For each keyword argument, it's value must be a list or tuple of the
        allowed argument values.

    Examples
    --------
    >>> @gluonnlp.data.register(segment=['train', 'test', 'dev'])
    ... class MyDataset(gluon.data.Dataset):
    ...     def __init__(self, segment='train'):
    ...         pass
    >>> my_dataset = gluonnlp.data.create('MyDataset')
    >>> print(type(my_dataset))
    <class 'gluonnlp.data.registry.MyDataset'>

    """

    def _real_register(class_):
        # Assert that the passed kwargs are meaningful
        for kwarg_name, values in kwargs.items():
            try:
                real_args = inspect.getfullargspec(class_).args
            except AttributeError:
                # pylint: disable=deprecated-method
                real_args = inspect.getargspec(class_.__init__).args

            if not kwarg_name in real_args:
                raise RuntimeError(
                    ('{} is not a valid argument for {}. '
                     'Only valid arguments can be registered.').format(
                         kwarg_name, class_.__name__))

            if not isinstance(values, (list, tuple)):
                raise RuntimeError(('{} should be a list of '
                                    'valid arguments for {}. ').format(
                                        values, kwarg_name))

        # Save the kwargs associated with this class_
        _REGSITRY_NAME_KWARGS[class_] = kwargs

        register_ = registry.get_register_func(Dataset, 'dataset')
        return register_(class_)

    if class_ is not None:
        # Decorator was called without arguments
        return _real_register(class_)

    return _real_register


def create(name, **kwargs):
    """Creates an instance of a registered dataset.

    Parameters
    ----------
    name : str
        The dataset name (case-insensitive).

    Returns
    -------
    An instance of :class:`mxnet.gluon.data.Dataset` constructed with the
    keyword arguments passed to the create function.

    """
    create_ = registry.get_create_func(Dataset, 'dataset')
    return create_(name, **kwargs)


def list_datasets(name=None):
    """Get valid datasets and registered parameters.

    Parameters
    ----------
    name : str or None, default None
        Return names and registered parameters of registered datasets. If name
        is specified, only registered parameters of the respective dataset are
        returned.

    Returns
    -------
    dict:
        A dict of all the valid keyword parameters names for the specified
        dataset. If name is set to None, returns a dict mapping each valid name
        to its respective keyword parameter dict. The valid names can be
        plugged in `gluonnlp.model.word_evaluation_model.create(name)`.

    """
    reg = registry.get_registry(Dataset)

    if name is not None:
        class_ = reg[name.lower()]
        return _REGSITRY_NAME_KWARGS[class_]
    else:
        return {
            dataset_name: _REGSITRY_NAME_KWARGS[class_]
            for dataset_name, class_ in registry.get_registry(Dataset).items()
        }



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""Encoder and decoder used in sequence-to-sequence learning."""
__all__ = ['Seq2SeqEncoder']

import mxnet as mx
from mxnet.gluon.block import Block

def _nested_sequence_last(data, valid_length):
    """

    Parameters
    ----------
    data : nested container of NDArrays/Symbols
        The input data. Each element will have shape (batch_size, ...)
    valid_length : NDArray or Symbol
        Valid length of the sequences. Shape (batch_size,)
    Returns
    -------
    data_last: nested container of NDArrays/Symbols
        The last valid element in the sequence.
    """
    assert isinstance(data, list)
    if isinstance(data[0], (mx.sym.Symbol, mx.nd.NDArray)):
        F = mx.sym if isinstance(data[0], mx.sym.Symbol) else mx.ndarray
        return F.SequenceLast(F.stack(*data, axis=0),
                              sequence_length=valid_length,
                              use_sequence_length=True)
    elif isinstance(data[0], list):
        ret = []
        for i in range(len(data[0])):
            ret.append(_nested_sequence_last([ele[i] for ele in data], valid_length))
        return ret
    else:
        raise NotImplementedError


class Seq2SeqEncoder(Block):
    r"""Base class of the encoders in sequence to sequence learning models.
    """

    def __call__(self, inputs, valid_length=None, states=None):  #pylint: disable=arguments-differ
        """Encode the input sequence.

        Parameters
        ----------
        inputs : NDArray
            The input sequence, Shape (batch_size, length, C_in).
        valid_length : NDArray or None, default None
            The valid length of the input sequence, Shape (batch_size,). This is used when the
            input sequences are padded. If set to None, all elements in the sequence are used.
        states : list of NDArrays or None, default None
            List that contains the initial states of the encoder.

        Returns
        -------
        outputs : list
            Outputs of the encoder.
        """
        return super(Seq2SeqEncoder, self).__call__(inputs, valid_length, states)

    def forward(self, inputs, valid_length=None, states=None):  #pylint: disable=arguments-differ
        raise NotImplementedError


class Seq2SeqDecoder(Block):
    """Base class of the decoders for sequence to sequence learning models.

    Given the inputs and the context computed by the encoder, generate the new
    states. Used in the training phase where we set the inputs to be the target
    sequence.

    """

    def init_state_from_encoder(self, encoder_outputs, encoder_valid_length=None):
        r"""Generates the initial decoder states based on the encoder outputs.

        Parameters
        ----------
        encoder_outputs : list of NDArrays
        encoder_valid_length : NDArray or None

        Returns
        -------
        decoder_states : list
        """
        raise NotImplementedError

    def forward(self, step_input, states, valid_length=None):  #pylint: disable=arguments-differ
        """Given the inputs and the context computed by the encoder, generate the new states.

        Used in the training phase where we set the inputs to be the target
        sequence.

        Parameters
        ----------
        inputs : NDArray
            The input embeddings. Shape (batch_size, length, C_in)
        states : list
            The initial states of the decoder.
        valid_length : NDArray or None
            valid length of the inputs. Shape (batch_size,)
        Returns
        -------
        output : NDArray
            The output of the decoder. Shape is (batch_size, length, C_out)
        states: list
            The new states of the decoder
        additional_outputs : list
            Additional outputs of the decoder, e.g, the attention weights

        """
        raise NotImplementedError


class Seq2SeqOneStepDecoder(Block):
    r"""Base class of the decoders in sequence to sequence learning models.

    In the forward function, it generates the one-step-ahead decoding output.

    """

    def init_state_from_encoder(self, encoder_outputs, encoder_valid_length=None):
        r"""Generates the initial decoder states based on the encoder outputs.

        Parameters
        ----------
        encoder_outputs : list of NDArrays
        encoder_valid_length : NDArray or None

        Returns
        -------
        decoder_states : list
        """
        raise NotImplementedError

    def forward(self, step_input, states):  #pylint: disable=arguments-differ
        """One-step decoding of the input

        Parameters
        ----------
        step_input : NDArray
            Shape (batch_size, C_in)
        states : list
            The previous states of the decoder
        Returns
        -------
        step_output : NDArray
            Shape (batch_size, C_out)
        states : list
        step_additional_outputs : list
            Additional outputs of the step, e.g, the attention weights
        """
        raise NotImplementedError



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=abstract-method
"""Helper functions."""

import os

__all__ = ['numba_njit', 'numba_prange', 'numba_jitclass', 'numba_types', 'get_home_dir']

try:
    from numba import njit, prange, jitclass, types  # pytype: disable=import-error
    numba_njit = njit(nogil=True)
    numba_prange = prange
    numba_jitclass = jitclass
    numba_types = types
except ImportError:
    # Define numba shims
    def identity(f):
        return f

    def numba_jitclass(spec):
        # pylint: disable=unused-argument
        return identity

    class NumbaTypes:
        """Shim for numba.types"""
        class NumbaType:
            """Shim for numba.types.type"""
            def __getitem__(self, x):
                # pylint: disable=unused-argument
                pass

            def __call__(self, *args, **kwargs):
                # pylint: disable=unused-argument
                pass

        def __getattr__(self, attr):
            return self.NumbaType()

    numba_types = NumbaTypes()
    numba_njit = identity
    numba_prange = range


def get_home_dir():
    """Get home directory for storing datasets/models/pre-trained word embeddings"""
    _home_dir = os.environ.get('MXNET_HOME', os.path.join('~', '.mxnet'))
    # expand ~ to actual path
    _home_dir = os.path.expanduser(_home_dir)
    return _home_dir



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
"""DataLoader. An extension of Gluon data loader that allows multi-shard sampling."""
__all__ = ['ShardedDataLoader']

import io
import pickle
import multiprocessing
from multiprocessing.pool import ThreadPool
from mxnet import context
from mxnet.gluon.data.dataloader import ForkingPickler, _as_in_context
from mxnet.gluon.data.dataloader import default_mp_batchify_fn, default_batchify_fn
from mxnet.gluon.data import sampler as _sampler

_worker_dataset = None
def _worker_initializer(dataset):
    """Initializer for processing pool."""
    # global dataset is per-process based and only available in worker processes
    # this is only necessary to handle MXIndexedRecordIO because otherwise dataset
    # can be passed as argument
    global _worker_dataset
    _worker_dataset = dataset

def _worker_fn(samples, batchify_fn, dataset=None):
    """Function for processing data in worker process."""
    # pylint: disable=unused-argument
    # it is required that each worker process has to fork a new MXIndexedRecordIO handle
    # preserving dataset as global variable can save tons of overhead and is safe in new process
    global _worker_dataset
    if isinstance(samples[0], (list, tuple)):
        batch = [batchify_fn([_worker_dataset[i] for i in shard]) for shard in samples]
    else:
        batch = batchify_fn([_worker_dataset[i] for i in samples])
    buf = io.BytesIO()
    ForkingPickler(buf, pickle.HIGHEST_PROTOCOL).dump(batch)
    return buf.getvalue()

def _thread_worker_fn(samples, batchify_fn, dataset):
    """Threadpool worker function for processing data."""
    if isinstance(samples[0], (list, tuple)):
        batch = [batchify_fn([dataset[i] for i in shard]) for shard in samples]
    else:
        batch = batchify_fn([dataset[i] for i in samples])
    return batch

class _MultiWorkerIter:
    """Internal multi-worker iterator for DataLoader."""
    def __init__(self, worker_pool, batchify_fn, batch_sampler, pin_memory=False,
                 worker_fn=_worker_fn, prefetch=0, dataset=None):
        self._worker_pool = worker_pool
        self._batchify_fn = batchify_fn
        self._batch_sampler = batch_sampler
        self._data_buffer = {}
        self._rcvd_idx = 0
        self._sent_idx = 0
        self._iter = iter(self._batch_sampler)
        self._worker_fn = worker_fn
        self._pin_memory = pin_memory
        self._dataset = dataset
        # pre-fetch
        for _ in range(prefetch):
            self._push_next()

    def __len__(self):
        return len(self._batch_sampler)

    def _push_next(self):
        """Assign next batch workload to workers."""
        r = next(self._iter, None)
        if r is None:
            return
        async_ret = self._worker_pool.apply_async(
            self._worker_fn, (r, self._batchify_fn, self._dataset))
        self._data_buffer[self._sent_idx] = async_ret
        self._sent_idx += 1

    def __next__(self):
        self._push_next()
        if self._rcvd_idx == self._sent_idx:
            assert not self._data_buffer, 'Data buffer should be empty at this moment'
            raise StopIteration

        assert self._rcvd_idx < self._sent_idx, 'rcvd_idx must be smaller than sent_idx'
        assert self._rcvd_idx in self._data_buffer, 'fatal error with _push_next, rcvd_idx missing'
        ret = self._data_buffer.pop(self._rcvd_idx)
        batch = pickle.loads(ret.get()) if self._dataset is None else ret.get()
        if self._pin_memory:
            batch = _as_in_context(batch, context.cpu_pinned())
        self._rcvd_idx += 1
        return batch

    def next(self):
        return self.__next__()

    def __iter__(self):
        return self


class ShardedDataLoader:
    """Loads data from a dataset and returns mini-batches of data.

    Parameters
    ----------
    dataset : Dataset
        Source dataset. Note that numpy and mxnet arrays can be directly used
        as a Dataset.
    batch_size : int
        Size of mini-batch.
    shuffle : bool
        Whether to shuffle the samples.
    sampler : Sampler
        The sampler to use. Either specify sampler or shuffle, not both.
    last_batch : {'keep', 'discard', 'rollover'}
        How to handle the last batch if batch_size does not evenly divide
        `len(dataset)`.

        keep - A batch with less samples than previous batches is returned.
        discard - The last batch is discarded if its incomplete.
        rollover - The remaining samples are rolled over to the next epoch.
    batch_sampler : Sampler
        A sampler that returns mini-batches. Do not specify batch_size,
        shuffle, sampler, and last_batch if batch_sampler is specified.
    batchify_fn : callable
        Callback function to allow users to specify how to merge samples
        into a batch. Defaults to `default_batchify_fn`::

            def default_batchify_fn(data):
                if isinstance(data[0], nd.NDArray):
                    return nd.stack(*data)
                elif isinstance(data[0], tuple):
                    data = zip(*data)
                    return [default_batchify_fn(i) for i in data]
                else:
                    data = np.asarray(data)
                    return nd.array(data, dtype=data.dtype)

    num_workers : int, default 0
        The number of multiprocessing workers to use for data preprocessing.
        `num_workers > 0` is not supported on Windows yet.
    pin_memory : boolean, default False
        If ``True``, the dataloader will copy NDArrays into pinned memory
        before returning them. Copying from CPU pinned memory to GPU is faster
        than from normal CPU memory.
    prefetch : int, default is `num_workers * 2`
        The number of prefetching batches only works if `num_workers` > 0.
        If `prefetch` > 0, it allow worker process to prefetch certain batches before
        acquiring data from iterators.
        Note that using large prefetching batch will provide smoother bootstrapping performance,
        but will consume more shared_memory. Using smaller number may forfeit the purpose of using
        multiple worker processes, try reduce `num_workers` in this case.
        By default it defaults to `num_workers * 2`.
    thread_pool : bool, default False
        If ``True``, use threading pool instead of multiprocessing pool. Using threadpool
        can avoid shared memory usage. If `DataLoader` is more IO bounded or GIL is not a killing
        problem, threadpool version may achieve better performance than multiprocessing.

    """
    def __init__(self, dataset, batch_size=None, shuffle=False, sampler=None,
                 last_batch=None, batch_sampler=None, batchify_fn=None,
                 num_workers=0, pin_memory=False, prefetch=None, thread_pool=False):
        self._dataset = dataset
        self._pin_memory = pin_memory
        self._thread_pool = thread_pool

        if batch_sampler is None:
            if batch_size is None:
                raise ValueError('batch_size must be specified unless ' \
                                 'batch_sampler is specified')
            if sampler is None:
                if shuffle:
                    sampler = _sampler.RandomSampler(len(dataset))
                else:
                    sampler = _sampler.SequentialSampler(len(dataset))
            elif shuffle:
                raise ValueError('shuffle must not be specified if sampler is specified')

            batch_sampler = _sampler.BatchSampler(
                sampler, batch_size, last_batch if last_batch else 'keep')
        elif batch_size is not None or shuffle or sampler is not None or \
                last_batch is not None:
            raise ValueError('batch_size, shuffle, sampler and last_batch must ' \
                             'not be specified if batch_sampler is specified.')

        self._batch_sampler = batch_sampler
        self._num_workers = num_workers if num_workers >= 0 else 0
        self._worker_pool = None
        self._prefetch = max(0, int(prefetch) if prefetch is not None else 2 * self._num_workers)
        if self._num_workers > 0:
            if self._thread_pool:
                self._worker_pool = ThreadPool(self._num_workers)
            else:
                self._worker_pool = multiprocessing.Pool(
                    self._num_workers, initializer=_worker_initializer, initargs=[self._dataset])
        if batchify_fn is None:
            if num_workers > 0:
                self._batchify_fn = default_mp_batchify_fn
            else:
                self._batchify_fn = default_batchify_fn
        else:
            self._batchify_fn = batchify_fn

    def __iter__(self):
        if self._num_workers == 0:
            def _same_process_iter():
                for batch in self._batch_sampler:
                    if isinstance(batch[0], (list, tuple)):
                        rets = [self._batchify_fn([self._dataset[idx] for idx in shard])
                                for shard in batch]
                        if self._pin_memory:
                            rets = [_as_in_context(ret, context.cpu_pinned()) for ret in rets]
                        yield rets
                    else:
                        ret = self._batchify_fn([self._dataset[idx] for idx in batch])
                        if self._pin_memory:
                            ret = _as_in_context(ret, context.cpu_pinned())
                        yield ret
            return _same_process_iter()

        # multi-worker
        return _MultiWorkerIter(self._worker_pool, self._batchify_fn, self._batch_sampler,
                                pin_memory=self._pin_memory,
                                worker_fn=_thread_worker_fn if self._thread_pool else _worker_fn,
                                prefetch=self._prefetch,
                                dataset=self._dataset if self._thread_pool else None)

    def __len__(self):
        return len(self._batch_sampler)

    def __del__(self):
        if self._worker_pool:
            # manually terminate due to a bug that pool is not automatically terminated
            # https://bugs.python.org/issue34172
            assert isinstance(self._worker_pool, multiprocessing.pool.Pool)
            self._worker_pool.terminate()



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""Convolutional encoder."""


__all__ = ['ConvolutionalEncoder']

from mxnet import gluon
from mxnet.gluon import nn
from ..initializer import HighwayBias

from .highway import Highway


class ConvolutionalEncoder(gluon.HybridBlock):
    r"""Convolutional encoder.

    We implement the convolutional encoder proposed in the following work::

        @inproceedings{kim2016character,
         title={Character-Aware Neural Language Models.},
         author={Kim, Yoon and Jernite, Yacine and Sontag, David and Rush, Alexander M},
         booktitle={AAAI},
         pages={2741--2749},
         year={2016}
        }

    Parameters
    ----------
    embed_size : int, default 15
        The input dimension to the encoder.
        We set the default according to the original work's experiments
        on PTB dataset with Char-small model setting.
    num_filters: Tuple[int], default (25, 50, 75, 100, 125, 150)
        The output dimension for each convolutional layer according to the filter sizes,
        which are the number of the filters learned by the layers.
        We set the default according to the original work's experiments
        on PTB dataset with Char-small model setting.
    ngram_filter_sizes: Tuple[int], default (1, 2, 3, 4, 5, 6)
        The size of each convolutional layer,
        and len(ngram_filter_sizes) equals to the number of convolutional layers.
        We set the default according to the original work's experiments
        on PTB dataset with Char-small model setting.
    conv_layer_activation: str, default 'tanh'
        Activation function to be used after convolutional layer.
        We set the default according to the original work's experiments
        on PTB dataset with Char-small model setting.
    num_highway: int, default '1'
        The number of layers of the Highway layer.
        We set the default according to the original work's experiments
        on PTB dataset with Char-small model setting.
    highway_layer_activation: str, default 'relu'
        Activation function to be used after highway layer.
        If you don't specify anything, no activation is applied
        (ie. "linear" activation: `a(x) = x`).
        We set the default according to the original work's experiments
        on PTB dataset with Char-small model setting.
    highway_bias : HighwayBias,
        default HighwayBias(nonlinear_transform_bias=0.0, transform_gate_bias=-2.0)
        The biases applied to the highway layer.
        We set the default according to the above original work.
    output_size: int, default None
        The output dimension after conducting the convolutions and max pooling,
        and applying highways, as well as linear projection.

    """
    def __init__(self,
                 embed_size=15,
                 num_filters=(25, 50, 75, 100, 125, 150),
                 ngram_filter_sizes=(1, 2, 3, 4, 5, 6),
                 conv_layer_activation='tanh',
                 num_highway=1,
                 highway_layer_activation='relu',
                 highway_bias=HighwayBias(nonlinear_transform_bias=0.0, transform_gate_bias=-2.0),
                 output_size=None,
                 **kwargs):
        super(ConvolutionalEncoder, self).__init__(**kwargs)

        self._embed_size = embed_size
        self._num_filters = num_filters
        self._ngram_filter_sizes = ngram_filter_sizes
        self._num_highway = num_highway
        self._output_size = output_size

        with self.name_scope():
            self._convs = gluon.contrib.nn.HybridConcurrent()
            maxpool_output_size = 0
            with self._convs.name_scope():
                for num_filter, ngram_size in zip(self._num_filters, self._ngram_filter_sizes):
                    seq = nn.HybridSequential()
                    seq.add(nn.Conv1D(in_channels=self._embed_size,
                                      channels=num_filter,
                                      kernel_size=ngram_size,
                                      use_bias=True))
                    seq.add(gluon.nn.HybridLambda(lambda F, x: F.max(x, axis=2)))
                    if conv_layer_activation is not None:
                        seq.add(nn.Activation(conv_layer_activation))
                    self._convs.add(seq)
                    maxpool_output_size += num_filter

            if self._num_highway:
                self._highways = Highway(maxpool_output_size,
                                         self._num_highway,
                                         activation=highway_layer_activation,
                                         highway_bias=highway_bias)
            else:
                self._highways = None
            if self._output_size:
                self._projection = nn.Dense(in_units=maxpool_output_size,
                                            units=self._output_size,
                                            use_bias=True)
            else:
                self._projection = None
                self._output_size = maxpool_output_size

    def hybrid_forward(self, F, inputs, mask=None): # pylint: disable=arguments-differ
        r"""
        Forward computation for char_encoder

        Parameters
        ----------
        inputs: NDArray
            The input tensor is of shape `(seq_len, batch_size, embedding_size)` TNC.
        mask: NDArray
            The mask applied to the input of shape `(seq_len, batch_size)`, the mask will
            be broadcasted along the embedding dimension.

        Returns
        ----------
        output: NDArray
            The output of the encoder with shape `(batch_size, output_size)`

        """
        if mask is not None:
            inputs = F.broadcast_mul(inputs, mask.expand_dims(-1))

        inputs = F.transpose(inputs, axes=(1, 2, 0))

        output = self._convs(inputs)

        if self._highways:
            output = self._highways(output)

        if self._projection:
            output = self._projection(output)

        return output



# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

# pylint: disable=wildcard-import
"""NLP toolkit."""

import warnings
import mxnet

from . import loss
from . import data
from . import embedding
from . import model
from . import metric
from . import utils
from . import vocab
from . import optimizer
from . import initializer
from .vocab import Vocab

__version__ = '0.9.2'

__all__ = ['data',
           'model',
           'embedding',
           'Vocab',
           'vocab',
           'loss',
           'initializer',
           'optimizer',
           'utils',
           'metric']

warnings.filterwarnings(module='gluonnlp', action='default', category=DeprecationWarning)
utils.version.check_version('1.6.0', warning_only=True, library=mxnet)
